#!/usr/bin/env python

from __future__ import print_function
from __future__ import division
from builtins import range
from builtins import object
from past.utils import old_div
import katropton as m
import tbox
import tbox.gmsh as gmsh
import shlex, subprocess
import os
import numpy as np

from fwk.wutils import parseargs

from PyTrilinos import Teuchos

import katropton.utilities as u

defaut_params = {
  'Correlation length X' : 20.,
  'Correlation length Y' : 20.,
  'Correlation length Z' : 0.,
  'Number of KL Terms X' : 5,
  'Number of KL Terms Y' : 5,
  'Number of KL Terms Z' : 3,
  'Dispersion level' : 0.1,
  'gamma' : 4368992.,
  'MueLu xml from current_dir' : True,
  'MueLu xml use template' : True,
  'MueLu xml file name' : '/SIMPLE_gs.xml',
  'MueLu xml template file name' : '/../preconditioners/SIMPLE_gs.xml.in',
  'MueLu xml params name' : ['$BLOCK_00_STRIDING_INFO','$BLOCK_11_STRIDING_INFO', \
                   '$BLOCK_00_GS_sweeps','$BLOCK_11_GS_sweeps', \
                   '$BLOCK_00_GS_damping','$BLOCK_11_GS_damping', \
                   '$SCHUR_omega','$SimpleSmoother_damping'],
  'MueLu xml params' : ['3','3','10','10','1.6','1.6','0.5','1.'],
  'Write matrix and vectors' : False,
  'Write txt files' : False
}

def Compute_QoIs(filename):
  import tboxVtk.reader as vtkR
  import tboxVtk.lineExtractor as vtkC

  QoI = np.zeros((4,))
  reader = vtkR.Reader()
  reader.open(filename)
  cutter = vtkC.LineExtractor(reader.reader.GetOutputPort())
  c_output = cutter.cut(1, [5.,5.,0.5,],[10.,5.,0.5,], tag_name='volume_id')
  pts, tri, vals = cutter.extract(c_output, ['von Mises'], atPoint = False)
  QoI[0] = vals['von Mises'][0]
  
  c_output = cutter.cut(1, [5.,5.,0.5,],[5.,10.,0.5,], tag_name='volume_id')
  pts, tri, vals = cutter.extract(c_output, ['von Mises'], atPoint = False)
  QoI[1] = vals['von Mises'][0]
  
  c_output = cutter.cut(1, [5.,5.,0.5,],[0.,5.,0.5,], tag_name='volume_id')
  pts, tri, vals = cutter.extract(c_output, ['von Mises'], atPoint = False)
  QoI[2] = vals['von Mises'][0]

  c_output = cutter.cut(1, [5.,5.,0.5,],[5.,0.,0.5,], tag_name='volume_id')
  pts, tri, vals = cutter.extract(c_output, ['von Mises'], atPoint = False)
  QoI[3] = vals['von Mises'][0]
  return QoI

class plate_model(object):
  def __init__(self, ensemble_size, randomVariable, msh, comm, file_dir, current_dir, p=defaut_params):
    
    if p['MueLu xml use template']:
      old_strings = p['MueLu xml params name']
      new_strings = p['MueLu xml params']

      muelu_xml_in = file_dir  + p['MueLu xml template file name']
      u.replace_strings(muelu_xml_in,current_dir+p['MueLu xml file name'],old_strings,new_strings)
    
    pbl = m.Problem(msh,comm)

    # Units are in mm
    E = 70000 * np.ones(ensemble_size) # 70e9 Pa = 70000 MPa
    nu = 0.35 * np.ones(ensemble_size)
    k = 0.
    d = 0.
    m.Medium(pbl, "body", "test",E,nu,k,d)

    m.uDirichlet(pbl, "clamped xyz", "Clamped",1,0.,1,0.,1,0.,0,0.,ensemble_size)
    m.uDirichlet(pbl, "clamped xy" , "Clamped",1,0.,1,0.,0,0.,0,0.,ensemble_size)
    m.uDirichlet(pbl, "clamped y"  , "Clamped",0,0.,1,0.,0,0.,0,0.,ensemble_size)

    norm = tbox.Pt(0,0,-1)
    cont = m.Contact(pbl,'inner part boundary','contact',norm)
    cont.setMeshTying()
    cont.setMaster(pbl,'outer part boundary',norm)
    f = 700 * np.ones(ensemble_size) # 700e6 Pa = 700 MPa
    zero = np.zeros(ensemble_size)
    one = np.ones(ensemble_size)

    m.Neumann(pbl, 'load 1','load',0,zero,1,f,0,zero,0,one,ensemble_size)
    m.Neumann(pbl, 'load 2','load',0,zero,1,-f,0,zero,0,one,ensemble_size)

    solverList = Teuchos.ParameterList()
    solverList['Ensemble Convergence Tolerance'] = 10**(-7)
    solverList['Maximum Iterations'] = 1000
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    if p['MueLu xml from current_dir']:
      solverList['MueLu xml file name'] = current_dir+p['MueLu xml file name']
    else:
      solverList['MueLu xml file name'] = file_dir+p['MueLu xml file name']
    solverList['convert MueLu xml file'] = True
    solverList['Print Teuchos timers'] = True

    solverList['gamma'] = p['gamma']
    solverList['Use preconditioner'] = True

    solverList['type']= 'BlockGmres'
    solverList['Maximum active set iteration'] = 1

    solverList['Write matrix and vectors'] = p['Write matrix and vectors']
    solverList['Write txt files'] = p['Write txt files']
    
    
    num_random_variables = randomVariable.shape[1]
    
    for i in range(0,num_random_variables):
        rand = randomVariable[:,i]
        m.RandomVariable(pbl,rand)

    randomParams = Teuchos.ParameterList()
    
    randomParams['Is random'] = True
    randomParams['Begin X'] = 0.
    randomParams['Begin Y'] = 0.
    randomParams['Begin Z'] = 0.
    randomParams['End X'] = 10.
    randomParams['End Y'] = 10.
    randomParams['End Z'] = 1.
    randomParams['Number random variables'] = num_random_variables
    randomParams['Correlation length X'] = p['Correlation length X']
    randomParams['Correlation length Y'] = p['Correlation length Y']
    randomParams['Correlation length Z'] = p['Correlation length Z']
    randomParams['Number of KL Terms X'] = p['Number of KL Terms X']
    randomParams['Number of KL Terms Y'] = p['Number of KL Terms Y']
    randomParams['Number of KL Terms Z'] = p['Number of KL Terms Z']
    randomParams['Maximum Nonlinear Solver Iterations'] = 5000
    randomParams['Bound Perturbation Size'] = 1E-10

    randomParams['Mean'] = old_div(E[0],(2.*(1.+nu[0])))
    randomParams['Dispersion level'] = p['Dispersion level']
    
    solverList['randomParams'] = randomParams
    
    useEP = (ensemble_size>1)
    
    self.pbl = pbl
    self.solverList = solverList
    self.ensemble_size = ensemble_size
    self.useEP = useEP

  def run(self,k=1):
    k = int(k)
    self.slv = m.IterativeSolver(self.pbl,k,self.solverList,3, self.useEP, self.ensemble_size)
    self.slv.start()
    f = open('timers.txt', 'w')
    print(self.slv.getTimers(), file=f)
    f.closed

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print("info: MPI found")
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print("info: MPI not found => MPI disabled")


    args=parseargs()

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    # Units in the meshes are in mm
    msh = u.fuse_meshes('plate_1.msh','plate_1_tmp','plate_2.msh','plate_2_tmp','plate_fused',file_dir,work_dir)

    np.random.seed(42)
    N = 1
    num_random_variables = 20
    #randomVariable = np.random.normal(0, 1, num_random_variables*N)
    #randomVariable = np.resize(randomVariable,(N,num_random_variables))
    randomVariable = np.zeros((N,num_random_variables))
    
    p = defaut_params
    p['Write txt files'] = True

    ensemble_size = 1
    
    pm = plate_model(ensemble_size, randomVariable, msh, comm, file_dir, work_dir, p)
    pm.run(args.k)

if __name__ == "__main__":
    main()
