#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div
from builtins import object
import numpy as np

from scipy.linalg import eig

tol = 1e-6

class Gauss_Seidel(object):
  def __init__(self, A,n=1,damping=1.):
    from scipy.sparse import tril
    from scipy.sparse import triu
    from scipy.sparse import eye
    from scipy.sparse import diags
    
    D = diags(A.diagonal(), 0)
    E = -tril(A,-1)
    F = -triu(A,1)
    
    self.x = np.zeros((A.shape[0],))
    
    self.M = (D-damping*E).tocsr()
    self.N = damping*F+(1-damping)*D
    self.damping = damping
    self.n = n
  
    self.shape = [A.shape[0],A.shape[0]]

  def dot(self,b):
    from scipy.sparse.linalg import spsolve_triangular
    self.x = spsolve_triangular(self.M, self.damping*b)
    for i in range (0,self.n-1):
      self.x = spsolve_triangular(self.M,self.N.dot(self.x)+self.damping*b)
    return self.x

class SIMPLE_preconditioned_matrix(object):
  def __init__(self, Q,invQ,G,GT,R,invR,omega=1.):
    from scipy.sparse import diags
    self.n1 = Q.shape[0]
    self.n2 = G.shape[1]
    
    self.x = np.zeros((self.n1+self.n2,))
    self.x_tmp = np.zeros((self.n1+self.n2,))

    self.omega = omega
    
    self.Q = Q
    self.G = G
    self.GT = GT
    
    self.R = R
  
    self.invQ = invQ
    self.invR = invR
    
    Qdiag = Q.diagonal()
    self.invD = diags(old_div(1,Qdiag), 0)
  
    self.shape = [Q.shape[0]+G.shape[1],Q.shape[0]+G.shape[1]]
  
  def dot(self,b):
    n1 = self.n1
    n2 = self.n2
    
    self.x[0:n1] = self.invQ.dot(b[0:n1])
    self.x[n1:n1+n2] = self.invR.dot(b[n1:n1+n2]-self.GT.dot(self.x[0:n1]))
    
    self.x_tmp[0:n1] = self.x[0:n1] - self.omega*self.invD.dot(self.G.dot(self.x[n1:n1+n2]))
    self.x_tmp[n1:n1+n2] = self.omega*self.x[n1:n1+n2]
    
    self.x[0:n1] = self.Q.dot(self.x_tmp[0:n1]) + self.G.dot(self.x_tmp[n1:n1+n2])
    self.x[n1:n1+n2] = self.GT.dot(self.x_tmp[0:n1])
    
    return self.x

def read_mm(filename,ensemble_id=-1,is_sparse=True):
    from scipy.sparse import csr_matrix
    
    X_dim = np.genfromtxt(filename, skip_header=1, max_rows=1)
    m = X_dim[0].astype(int)
    n = X_dim[1].astype(int)
    
    if is_sparse:
      nnz = X_dim[2].astype(int)
      if ensemble_id==-1:
        #No ensemble typed data
        X = np.loadtxt(filename, skiprows=2)

        row = X[:,0].astype(int)-1
        col = X[:,1].astype(int)-1
        data = X[:,2]
      
      else:
        #Ensemble typed data
        X = np.genfromtxt(filename, skip_header=2, usecols = (0, 1, 3+ensemble_id))

        row = X[:,0].astype(int)-1
        col = X[:,1].astype(int)-1
        data = X[:,2]

      mask = data!=0.
      A = csr_matrix((data[mask], (row[mask], col[mask])), shape=(m, n))
      B = A.tocsc()
    else:
      X = np.loadtxt(filename, skiprows=2)
      if n!= 1:
        B = X.reshape((m,n)).T
      else:
        B = X.reshape((m,))
    
    return B
    
def create_block(A11,A12,A21,A22):
    from scipy.sparse import bmat
    A = bmat([[A11,A12], [A21,A22]])
    return A

def get_symmetrical_part(A):
    A2 = A
    tmp = A - A.T
    A2[tmp!=0] = 0
    return A2

def Schur_complement(GT,invQ,G):
    R = -GT*invQ*G
    return R

def approx_Schur_complement(Q,G,GT):
    from scipy.sparse import diags
    Qdiag = Q.diagonal()
    invD = diags(old_div(1,Qdiag), 0)
    R = Schur_complement(GT,invD,G)
    return R

def SIMPLE_prec(Q,invQ,G,GT,R,invR,use_invQG=False,omega=1.):
    from scipy.sparse import eye
    from scipy.sparse import diags
    Qdiag = Q.diagonal()
    invD = diags(old_div(1,Qdiag), 0)
    n1 = Q.shape[0]
    n2 = R.shape[0]
    if use_invQG:
      print('Warning: use_invQG is OK in SIMPLE only if we use a direct solver for each block for now')
      print('Warning: this returned the preconditioned matrix and not the preconditioner matrix')
      GTinvQ = invQ.T
      invP = create_block(eye(n1)-(eye(n1)-Q*invD)*G*invR*GTinvQ,(eye(n1)-Q*invD)*G*invR,None,eye(n2))
    else:
      B = create_block(eye(n1),-omega*invD*G,None,omega*eye(n2))
      invM = create_block(invQ,None,-invR*GT*invQ,invR)
      invP = B*invM

    return invP

def compute_inv(A):
    from scipy.sparse.linalg import inv
    invA = inv(A)
    return invA

def compute_inv_on_image(A,B):
    #Compute the inverse of A applied on the image of B:
    #invAB = A^{—1} * B
    from scipy.sparse.linalg import spsolve
    from scipy.sparse import issparse
    
    if issparse(B):
      B2 = B.todense()
    else:
      B2 = B

    invAB = spsolve(A,B2)
    return invAB
    
def compute_inv_GS(A,n,damping=1.):
    from scipy.sparse import tril
    from scipy.sparse import triu
    from scipy.sparse import eye
    from scipy.sparse import diags
    from scipy.sparse import issparse
    Adiag = A.diagonal()
    D = diags(Adiag, 0)
    E = -tril(A,-1)
    F = -triu(A,1)
    
    M = (D-damping*E).tocsc()
    N = damping*F+(1-damping)*D
    invM = compute_inv(M)
    invMN = invM*N
    invA = invM*damping
    
    invMN_pow = invMN
    
    for i in range (0,n-1):        
        invA = invA + invMN_pow*invM*damping
        invMN_pow = invMN_pow * invMN
    
    return invA

def compute_VW(A,b,m=0,N=1):
    from scipy.linalg import qr
    n = A.shape[0]
    
    if m==0:
      m = n
    
    V = np.zeros((n,m))
    W = np.zeros((n,m))

    tmp = b.reshape((n,))
    previous_vec = np.zeros((n,))
    V[:,0] = old_div(tmp,np.linalg.norm(tmp))
    previous_vec = A.dot(V[:,0])
    W[:,0] = old_div(previous_vec,np.linalg.norm(previous_vec))
    
    for i in range(1,m):
      tmp = previous_vec
      for j in range(0,i):
        tmp = tmp - (np.inner(tmp,V[:,j])) * V[:,j]
      norm = np.linalg.norm(tmp)
      if norm<=tol:
        m = i
        break
      tmp = old_div(tmp,norm)
      V[:,i] = tmp
      previous_vec = A.dot(V[:,i])
      norm = np.linalg.norm(previous_vec)
      W[:,i] = old_div(previous_vec,norm)
    V = V[:,0:m]
    W = W[:,0:m]
    
    q, r = np.linalg.qr(W)
    W = q
    
    for i_N in range(0,N):
      q, r = np.linalg.qr(V)
      V = q
      if i_N != 0:
        q, r = np.linalg.qr(W)
        W = q

    product = np.dot(W.T,b)

    for i in range(0,len(product)):
      if product[i] < 0:
        W[:,i] = -W[:,i]
    return V, W

def compute_V(A,b,m=0,N=0):
    n = A.shape[0]
    
    if m==0:
      m = n
    
    V = np.zeros((n,m))

    tmp = b.reshape((n,))
    V[:,0] = old_div(tmp,np.linalg.norm(tmp))
    for i in range(1,m):
      tmp = A.dot(V[:,i-1])
      for j in range(0,i):
        tmp = tmp - (np.inner(tmp,V[:,j])) * V[:,j]
      norm = np.linalg.norm(tmp)
      if norm<=tol:
        m = i
        break
      else:
        tmp = old_div(tmp,norm)
        V[:,i] = tmp
    V = V[:,0:m]
    for i in range(0,N):
      for i in range(1,m):
        tmp = V[:,i]
        for j in range(0,i):
          tmp = tmp - (np.inner(tmp,V[:,j])) * V[:,j]
        norm = np.linalg.norm(tmp)
        if norm<=tol:
          m = i
          break
        else:
          tmp = old_div(tmp,norm)
          V[:,i] = tmp
      V = V[:,0:m]
    return V

def compute_W(A,b,V,N=0):
    from scipy.linalg import qr
    n = V.shape[0]
    m = V.shape[1]
    tmp = A.dot(V[:,0])
    W = np.zeros((n,m))

    W[:,0] = old_div(tmp,np.linalg.norm(tmp))

    for i in range(1,m):
      tmp = A.dot(V[:,i])
      for j in range(0,i):
        tmp = tmp - (np.inner(tmp,W[:,j])) * W[:,j]
      norm = np.linalg.norm(tmp)
      W[:,i] = old_div(tmp,norm)
    for i in range(0,N):
      for i in range(1,m):
        tmp = W[:,i]
        for j in range(0,i):
          tmp = tmp - (np.inner(tmp,W[:,j])) * W[:,j]
        W[:,i] = old_div(tmp,np.linalg.norm(tmp))

    product = np.dot(W.T,b)

    for i in range(0,len(product)):
      if product[i] < 0:
        W[:,i] = -W[:,i]

    return W

def compute_eig_Q(V,W):
    Q = np.dot(V.T,W)
    a, Z = eig(Q)

    return a, Z

def compute_deltas_omegas(V,W,b):
    import cmath
    deltas, Z = compute_eig_Q(V,W)
    omegas = np.zeros((Z.shape[1],))
    for i in range(0,len(omegas)):
      tmp = cmath.polar(Z[0,i])
      omegas[i] = tmp[0]**2

    return deltas, omegas
    
def eigenvalues_ZE(Q,invQ,G,GT,R,invR,use_invQG=False):
    from scipy.sparse import diags
    from scipy.sparse import issparse
    Qdiag = Q.diagonal()
    D = diags(old_div(1,Qdiag), 0)
    invD = diags(old_div(1,Qdiag), 0)
    n1 = Q.shape[0]
    n2 = R.shape[0]
    v = np.ones((n1+n2,), dtype=complex)
    
    J = invD*(D-Q)
    if use_invQG:
      invQG = invQ
    else:
      invQG =invQ*G
    ZE = J*invQG*invR*GT
    
    if issparse(ZE):
      ZE2 = ZE.todense()
    else:
      ZE2 = ZE

    a, V = eig(ZE2)
    
    v[0:n1] = a
    
    return v

def eigenvalues_SR(invQ,G,GT,R,use_invQG=False):
    from scipy.sparse import issparse
    n1 = invQ.shape[0]
    n2 = R.shape[0]
    v = np.ones((n1+n2,), dtype=complex)

    if use_invQG:
      invQG = invQ
    else:
      invQG =invQ*G

    S = -GT*invQG
    
    if issparse(S):
      S2 = S.todense()
    else:
      S2 = S

    if issparse(R):
      R2 = R.todense()
    else:
      R2 = R

    a, V = eig(S2,R2)
    
    v[0:n2] = a
    
    return v

def compute_1_eigenvector_per_eigenvale(A,v):
    from scipy.sparse.linalg import eigs
    n = A.shape[0]
    if len(v) != 1:
      nv = v.shape[0]
    else:
      nv = 1

    V = np.zeros((n,nv))
    b = np.zeros((n,))
    
    for i in range(0,nv):
      values, vectors = eigs(A, k=1, sigma=v[i].real)
      V[:,i] = vectors.reshape((n,))
    
    return V

def jacobi_iteration_matrix(A):
    from scipy.sparse import diags
    from scipy.sparse import eye
    
    diagA = A.diagonal()
    invD = diags(old_div(1,diagA), 0)
    
    jacobi_iteration_matrix = eye(A.shape[0]) - np.dot(invD,A)
    return jacobi_iteration_matrix

def is_SPD(A,tol=1e-8):
    # Linked to theorem 10.1.2 (p. 512 in Golub, Matrix Computation)
    # if A is SPD, then GS converge for any initial guess
    A = A.todense()
    if not np.allclose(A, A.T, atol=tol): return False
    else:
        try:
            L = np.linalg.cholesky(A)
        except:
            return False
    return True

def compute_eig_and_departure_from_orthogonality(A):
    from numpy import linalg as LA
    w, v = LA.eig(A)
    cond = LA.cond(v)
    return w, v, cond

def diagonal_dominance(A):
    is_diagonal_dominance = True
    for i in range(0,A.shape[0]):
        diag = np.absolute(A[i,i])
        sum_row = np.sum(np.absolute(A[i,:]))
        sum_row = sum_row - diag
        if  diag <= sum_row:
          print(str(diag) + ' <= ' + str(sum_row) + ' at row ' + str(i))
          is_diagonal_dominance = False
          break
    return is_diagonal_dominance


def spectral_radius(A):
    from scipy.sparse.linalg import eigs
    w, v = eigs(A, k=10,which ='LM')
    
    spectral_radius_A = np.absolute(w[0])
    return spectral_radius_A

def sparse_matrix_2_norm(A):
    from scipy.sparse.linalg import svds
    u, s, vt = svds(A, k=1)
    norm_A = s[0]
    return norm_A

def compute_theta(v):
    import cmath
    theta = np.zeros((len(v),))
    for x in range(len(v)):
      tmp = cmath.polar(v[x])
      theta[x] = tmp[1]
    return theta

def compute_cum_delta(theta,deltas,omegas,n):
    theta_discrete = np.linspace(-np.pi,np.pi,n)
    cum_delta = np.zeros((n,))
    deriv_delta = np.zeros((n,))
    cum_delta_omega = np.zeros((n,))
    deriv_delta_omega = np.zeros((n,))
    
    n_theta = len(theta)
    
    first_index = 0
    second_index = 0
    for i in range(1,n):
      for j in range(second_index,n_theta):
        if theta_discrete[i] < theta[j]:
          second_index = j
          break
      #indices = np.argwhere((theta_discrete[i-1] <= theta ) & (theta < theta_discrete[i]))
      cum_delta[i] = cum_delta[i-1] + second_index-first_index
      cum_delta_omega[i] = cum_delta_omega[i-1] + np.sum(omegas[first_index:second_index])
      deriv_delta[i] = second_index-old_div(first_index, (2*np.pi/n))
      deriv_delta_omega[i] = old_div(np.sum(omegas[first_index:second_index]), (2*np.pi/n))
      first_index = second_index

    percent_delta =  old_div(cum_delta,cum_delta[n-1])
    percent_delta_omega = old_div(cum_delta_omega,cum_delta_omega[n-1])

    return theta_discrete, cum_delta, deriv_delta, cum_delta_omega, deriv_delta_omega, percent_delta, percent_delta_omega

def write_eigenvalues(v,filename):
    np.savetxt(filename, v, fmt='%f %f')

def write_eigenvectors(v,V,filename):
    np.savetxt(filename+'_eigenvalues.txt', v, fmt='%f %f')
    nv = V.shape[1]
    for i in range(0,nv):
      np.savetxt(filename+'_eigenvectors_'+str(i)+'.txt', V[:,i])

def plot_eigenvalues_on_unit_circle(v):
    import matplotlib.pyplot as plt
    import cmath
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='polar')
    an = np.linspace(0, 2*np.pi, 100)
    ax.plot(an, np.ones(an.shape), zorder=1)
    for x in range(len(v)):
      tmp = cmath.polar(v[x])
      r = tmp[0]
      theta = tmp[1]
      ax.plot(theta,r,'ro', zorder=2)
    ax.set_ylim(0,1.5)
    ax.grid(True)

def plot_eigenvalues_on_unit_circle_highlight(v,indices):
    import matplotlib.pyplot as plt
    import cmath
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='polar')
    an = np.linspace(0, 2*np.pi, 100)
    ax.plot(an, np.ones(an.shape), zorder=1)
    for x in range(len(v)):
      tmp = cmath.polar(v[x])
      r = tmp[0]
      theta = tmp[1]
      if indices[x] == 0:
        ax.plot(theta,r,'go',alpha=0.3, zorder=2)
      else:
        ax.plot(theta,r,'ro', zorder=10)
    ax.set_ylim(0,1.5)
    ax.grid(True)

def plot_polar_function(theta,f,rho_min,rho_max):
    import matplotlib.pyplot as plt
    scale = old_div((rho_max-rho_min), np.max(f))
    rho = rho_min + scale*f
    plt.plot(theta, rho, zorder=3)
    plt.ylim(0,rho_max+0.5)

def plot_omegas(v):
    import matplotlib.pyplot as plt
    plt.figure()
    v = -np.sort(-v)
    plt.semilogy(v)
    plt.grid(True)

def plot_omegas_percent(v):
    import matplotlib.pyplot as plt
    plt.figure()
    v = -np.sort(-v)
    plt.plot(old_div(np.cumsum(v),np.sum(v)))
    plt.grid(True)

def main():
    filename = 'ApCTC_mm.txt'
    Q = read_mm(filename)
    filename = 'B_mm.txt'
    GT = read_mm(filename)
    G =GT.transpose()

    A = create_block(Q,G,GT,None)
    
    R = approx_Schur_complement(Q,G,GT)
    
    invQ = compute_inv(Q)
    invR = compute_inv(R)
    invP = SIMPLE_prec(Q,invQ,G,GT,R,invR)

    AinvP = (A*invP).todense()

    a, v = eig(AinvP)

    display = 1
    
    if display:
        import matplotlib.pyplot as plt
        for x in range(len(a)):
            plt.plot(a[x].real,a[x].imag,'ro',label='python')
        limit=np.max(np.ceil(np.absolute(a)))
        plt.xlim((-limit,limit))
        plt.ylim((-limit,limit))
        plt.show()

    v2 = eigenvalues_SR(Q,invQ,G,GT,R,invR)
    
    write_eigenvalues(a,'eigenvalues_direct_direct.txt')
    write_eigenvalues(v2,'eigenvalues_direct_direct_2.txt')

if __name__ == "__main__":
    main()
