#ifndef WCONTACTS_LIST_H
#define WCONTACTS_LIST_H

#include "katropton.h"

#include "Map.h"
#include "Nodes_list.h"
#include "Elements_list.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Kokkos_ViewFactory.hpp>

namespace katropton
{
namespace discrete_problem
{
class Contacts_list
{
private:
        bool at_least_one_sticking = false;

        Kokkos::View<int **, Kokkos::LayoutRight> contacts;
        Kokkos::View<int **, Kokkos::LayoutRight> contacts_dirichlet;

        Kokkos::View<int *, Kokkos::LayoutRight> shape_function;

        Kokkos::View<int *, Kokkos::LayoutRight> signorini;

        Kokkos::View<int *, Kokkos::LayoutRight> initially_open;

        Kokkos::View<int *, Kokkos::LayoutRight> no_update;
        Kokkos::View<int *, Kokkos::LayoutRight> sticking;
        Kokkos::View<int *, Kokkos::LayoutRight> mesh_tying;

        Kokkos::View<double *, Kokkos::LayoutRight> contacts_scale;

        Kokkos::View<int **, Kokkos::LayoutRight> initially_closed_nodes;

        Kokkos::View<double **, Kokkos::LayoutRight> contacts_normal;

        Tpetra::Vector<>::local_ordinal_type contacts_number;
        Tpetra::Vector<>::local_ordinal_type contacts_slave_nodes_size;
        Tpetra::Vector<>::local_ordinal_type contacts_slave_elems_size;
        Tpetra::Vector<>::local_ordinal_type contacts_master_nodes_size;
        Tpetra::Vector<>::local_ordinal_type contacts_master_elems_size;
        Tpetra::Vector<>::local_ordinal_type contacts_slave_normal;
        Tpetra::Vector<>::local_ordinal_type contacts_master_normal;
        Tpetra::Vector<>::local_ordinal_type contacts_slave_nodes;
        Tpetra::Vector<>::local_ordinal_type contacts_slave_elems;
        Tpetra::Vector<>::local_ordinal_type contacts_master_nodes;
        Tpetra::Vector<>::local_ordinal_type contacts_master_elems;

public:
        Contacts_list();
        Contacts_list(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<Elements_list> elements_list,
                      Teuchos::RCP<Nodes_list> nodes_list);

        ~Contacts_list()
        {
                Kokkos::resize(contacts, 0, 0);
                Kokkos::resize(contacts_normal, 0, 0);
                Kokkos::resize(shape_function, 0);
                Kokkos::resize(signorini, 0);
                Kokkos::resize(initially_open, 0);
                Kokkos::resize(no_update, 0);
                Kokkos::resize(sticking, 0);
                Kokkos::resize(mesh_tying, 0);
                Kokkos::resize(contacts_scale, 0);
                Kokkos::resize(contacts_dirichlet, 0, 0);
                Kokkos::resize(initially_closed_nodes, 0, 0);
        }

        inline local_ordinal_type getContactNumber() { return contacts_number; };
        inline local_ordinal_type getSlaveNodesSize(local_ordinal_type i)
        {
                return contacts(i, contacts_slave_nodes_size);
        };
        inline local_ordinal_type getSlaveElementsSize(local_ordinal_type i)
        {
                return contacts(i, contacts_slave_elems_size);
        };
        inline local_ordinal_type getMasterNodesSize(local_ordinal_type i)
        {
                return contacts(i, contacts_master_nodes_size);
        };
        inline local_ordinal_type getMasterElementsSize(local_ordinal_type i)
        {
                return contacts(i, contacts_master_elems_size);
        };

        inline local_ordinal_type getSlaveNode(local_ordinal_type i, local_ordinal_type n1)
        {
                return contacts(i, contacts_slave_nodes + n1);
        };
        inline local_ordinal_type getSlaveElement(local_ordinal_type i, local_ordinal_type e1)
        {
                return contacts(i, contacts_slave_elems + e1);
        };
        inline local_ordinal_type getMasterNode(local_ordinal_type i, local_ordinal_type n1)
        {
                return contacts(i, contacts_master_nodes + n1);
        };
        inline local_ordinal_type getMasterElement(local_ordinal_type i, local_ordinal_type e1)
        {
                return contacts(i, contacts_master_elems + e1);
        };

        inline double getSlaveNormal(local_ordinal_type i, local_ordinal_type j)
        {
                return contacts_normal(i, j);
        }
        inline double getMasterNormal(local_ordinal_type i, local_ordinal_type j)
        {
                return contacts_normal(i, j + 3);
        }

        inline bool isStandardShapeFunction(local_ordinal_type i)
        {
                return shape_function(i) == 1;
        }

        inline bool isInitiallyOpen(local_ordinal_type i)
        {
                return initially_open(i) == 1;
        }

        inline bool isInitiallyClose(local_ordinal_type i)
        {
                return initially_open(i) == 0;
        }

        inline bool isUpdated(local_ordinal_type i)
        {
                return no_update(i) == 0;
        }

        inline bool isNotUpdated(local_ordinal_type i)
        {
                return no_update(i) == 1;
        }

        inline bool isSticking(local_ordinal_type i)
        {
                return sticking(i);
        }

        inline bool isTying(local_ordinal_type i)
        {
                return mesh_tying(i);
        }

        inline bool hasAtLeastOneSticking(void)
        {
                return at_least_one_sticking;
        }

        inline bool isSignorini(local_ordinal_type i)
        {
                return signorini(i) == 1;
        }

        inline bool isNotSignorini(local_ordinal_type i)
        {
                return signorini(i) == 0;
        }

        inline bool isSignorini()
        {
                for (local_ordinal_type i = 0; i < contacts_number; ++i)
                        if (isSignorini(i))
                                return true;
                return false;
        }

        inline bool isNotSignorini()
        {
                for (local_ordinal_type i = 0; i < contacts_number; ++i)
                        if (isSignorini(i))
                                return false;
                return true;
        }

        inline local_ordinal_type slaveDirichletSize(local_ordinal_type i)
        {
                return contacts_dirichlet(i, 0);
        }

        inline double getMortarScale(local_ordinal_type i)
        {
                return contacts_scale(i);
        }

        bool isASlaveNode(local_ordinal_type i, local_ordinal_type j)
        {
                local_ordinal_type size = getSlaveNodesSize(i);
                if (size == 0)
                        return false;
                for (auto j2 = 0; j2 < size; ++j2)
                        if (j == getSlaveNode(i, j2))
                                return true;
                return false;
        }

        local_ordinal_type getInterfaceOfSlaveNode(local_ordinal_type j)
        {
                //if (contacts_number == 1)
                //    return 0;

                for (local_ordinal_type i = 0; i < contacts_number; ++i)
                        if (isASlaveNode(i, j))
                                return i;
                return 0;
        }

        bool isASlaveDirichletNode(local_ordinal_type i, local_ordinal_type j)
        {
                local_ordinal_type size = slaveDirichletSize(i);
                if (size == 0)
                        return false;
                for (auto j2 = 0; j2 < size; ++j2)
                        if (j == contacts_dirichlet(i, j2 + 1))
                                return true;
                return false;
        }

        bool isNodeInitiallyClosed(local_ordinal_type i, local_ordinal_type j)
        {
                local_ordinal_type size = initially_closed_nodes(i, 0);
                if (size == 0)
                        return false;
                for (auto j2 = 0; j2 < size; ++j2)
                        if (j == initially_closed_nodes(i, j2 + 1))
                                return true;
                return false;
        }
};

}; // namespace discrete_problem
}; // namespace katropton

#endif //WCONTACTS_LIST_H
