//#define ElementMatrices_DEBUG

template <typename scalar, int element_type, int element_size>
ElementMatrices<scalar, element_type, element_size>::ElementMatrices(Teuchos::RCP<katropton::discrete_problem::Domain<scalar>> _domain, size_t _numPrimalDPN, size_t pool_size) : domain(_domain), numPrimalDPN(_numPrimalDPN)
{

    K = Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A", pool_size, 3 * element_size, 4 * element_size);
    L = Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A", pool_size, element_size, element_size);
    S = Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A", pool_size, 3 * element_size, element_size);

    K_e = Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A", pool_size, 3 * element_size, 3 * element_size);
    L_e = Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A", pool_size, element_size, element_size);
    S_e = Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A", pool_size, 3 * element_size, element_size);

    material = Kokkos::View<scalar **, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A", pool_size, 3 * element_size);
}

template <typename scalar, int element_type, int element_size>
void ElementMatrices<scalar, element_type, element_size>::compute(int e, int i_thread)
{

    tbox::Cache *cache;

    /* @todo could probably be refactored into cache = element->getVCache(); */
    if (element_type == static_cast<int>(tbox::ELTYPE::HEX8))
    {
        tbox::CacheHex8 &cachehex8 = this->trilinosHex8GetCache();
        cache = &cachehex8;
    }
    else if (element_type == static_cast<int>(tbox::ELTYPE::TETRA4))
    {
        tbox::CacheTetra4 &cachetetra4 = this->trilinosTetra4GetCache();
        cache = &cachetetra4;
    }

    std::vector<tbox::Pt> gauss_p = cache->getVGauss().p;
    std::vector<double> gauss_w = cache->getVGauss().w;

    tVector<scalar, element_size * 4> tValues;

    auto K_current = subview(K, i_thread, Kokkos::ALL(), Kokkos::ALL());
    auto L_current = subview(L, i_thread, Kokkos::ALL(), Kokkos::ALL());
    auto S_current = subview(S, i_thread, Kokkos::ALL(), Kokkos::ALL());

    auto K_e_current = subview(K_e, i_thread, Kokkos::ALL(), Kokkos::ALL());
    auto L_e_current = subview(L_e, i_thread, Kokkos::ALL(), Kokkos::ALL());
    auto S_e_current = subview(S_e, i_thread, Kokkos::ALL(), Kokkos::ALL());

    Kokkos::deep_copy(K_e_current, 0);
    Kokkos::deep_copy(L_e_current, 0);
    Kokkos::deep_copy(S_e_current, 0);

    auto element_material = domain->elements_list->Get_Element_Material(e);

    tVector<double, 3> factor1, factor2, dffi, dffj;
    tVector<scalar, 3> factor3;
    tMatrix<double, 3, 3> J, Jinv;
    tMatrix<scalar, 3, 3> TMP1;
    tMatrix<scalar, 9, 9> H;
    tVector<double, element_size> ff;
    scalar E, nu, lambda, G, conductivity, beta;
    tMatrix<scalar, 3, 3> K_conductivity;
    tMatrix<scalar, 3, 3> D;

    katropton::discrete_problem::Material<scalar> material = domain->materials_list->getMaterial(element_material, i_thread);
    if (domain->random_field->isRandom)
    {
        double x = 0., y = 0., z = 0.;
        for (auto j = 0; j < element_size; ++j)
        {
            x += domain->nodes_list->nodes(domain->elements_list->Get_Element_Node(e, j), 0) / element_size;
            y += domain->nodes_list->nodes(domain->elements_list->Get_Element_Node(e, j), 1) / element_size;
            z += domain->nodes_list->nodes(domain->elements_list->Get_Element_Node(e, j), 2) / element_size;
        }

        G = domain->random_field->operator()(x, y, z);
#ifdef ElementMatrices_DEBUG
        printf("G_0 = %f\n", EnsembleTraits<scalar>::coeff(G, 0));
#endif

        //E = material.getE();
        nu = material.getNu();
        lambda = 2. * G * nu / (1. - 2. * nu);
        //lambda = (nu*E)/((1+nu)*(1-2*nu));
    }
    else
    {
        E = material.getE();
        nu = material.getNu();
        lambda = (nu * E) / ((1 + nu) * (1 - 2 * nu));
        G = E / (2 * (1 + nu));
    }

    // Machine epsilon

    double epsilon = 2e-16;
    // Looping on gauss points
    for (auto a = 0; a < gauss_p.size(); ++a)
    {
        for (auto j = 0; j < element_size; ++j)
            if (element_size == 8)
                ff(j) = (static_cast<tbox::CacheHex8 *>(cache)->ff)[a][j];
            else if (element_size == 4)
                ff(j) = (static_cast<tbox::CacheTetra4 *>(cache)->ff)[a][j];

        // H matrix
        H.clean();

        for (auto i = 0; i < 3; ++i)
            for (auto j = 0; j < 3; ++j)
                for (auto k = 0; k < 3; ++k)
                    for (auto l = 0; l < 3; ++l)
                    {
                        if (i == j && k == l)
                            H(i * 3 + j, k * 3 + l) = lambda;
                        if (i == k && j == l)
                            H(i * 3 + j, k * 3 + l) = H(i * 3 + j, k * 3 + l) + G;
                        if (i == l && j == k)
                            H(i * 3 + j, k * 3 + l) = H(i * 3 + j, k * 3 + l) + G;
                    }

        conductivity = material.getK();
        K_conductivity.clean();
        for (auto i = 0; i < 3; ++i)
            K_conductivity(i, i) = conductivity;

        beta = material.getBeta();
        D.clean();
        for (auto i = 0; i < 3; ++i)
            D(i, i) = -3 * (lambda + (2. / 3.) * G) * beta;

        double detJ = this->buildJ(a, J, e, *(domain->elements_list), *(domain->nodes_list));

#ifdef ElementMatrices_DEBUG
        printf("detJ = %f\n", detJ);
#endif

        Jinv = J.inv();

        for (auto i = 0; i < element_size; ++i)
        {

            for (auto k = 0; k < 3; ++k)
                if (element_size == 8)
                    dffi(k) = (static_cast<tbox::CacheHex8 *>(cache)->dff)[a][i][k];
                else if (element_size == 4)
                    dffi(k) = (static_cast<tbox::CacheTetra4 *>(cache)->dff)[a][i][k];
            factor1 = Jinv * dffi;
#ifdef ElementMatrices_DEBUG
            printf("dffi = [%f, %f, %f]\n", dffi(0), dffi(1), dffi(2));
            printf("factor1 = [%f, %f, %f]\n", factor1(0), factor1(1), factor1(2));
#endif
            // K
            for (auto j = i; j < element_size; ++j)
            {
                for (auto k = 0; k < 3; ++k)
                    if (element_size == 8)
                        dffj(k) = (static_cast<tbox::CacheHex8 *>(cache)->dff)[a][j][k];
                    else if (element_size == 4)
                        dffj(k) = (static_cast<tbox::CacheTetra4 *>(cache)->dff)[a][j][k];
                factor2 = Jinv * dffj;

                TMP1.clean();
                for (auto k = 0; k < 3; ++k)
                    for (auto m = 0; m < 3; ++m)
                    {
                        for (auto l = 0; l < 3; ++l)
                            for (auto n = 0; n < 3; ++n)
                                TMP1(l, n) = H(k * 3 + l, m * 3 + n);

                        factor3 = TMP1 * factor2;

                        scalar dotProduct = factor1.dotproduct(factor3);
#ifdef ElementMatrices_DEBUG
                        printf("dotProduct_0 = %f\n", EnsembleTraits<scalar>::coeff(dotProduct, 0));
#endif
                        K_e_current(i * 3 + k, j * 3 + m) += dotProduct * gauss_w[a] * detJ;
                    }
            }

            // L
            factor3 = K_conductivity * factor1;
            for (auto j = i; j < element_size; ++j)
            {
                for (auto k = 0; k < 3; ++k)
                    if (element_size == 8)
                        dffj(k) = (static_cast<tbox::CacheHex8 *>(cache)->dff)[a][j][k];
                    else if (element_size == 4)
                        dffj(k) = (static_cast<tbox::CacheTetra4 *>(cache)->dff)[a][j][k];

                factor2 = Jinv * dffj;
                scalar dotProduct = factor3.dotproduct(factor2);
                L_e_current(i, j) += dotProduct * gauss_w[a] * detJ;
            }

            // S
            factor3 = D * factor1;

            for (auto j = 0; j < element_size; ++j)
                for (auto k = 0; k < 3; ++k)
                    S_e_current(i * 3 + k, j) += factor3(k) * ff(j) * gauss_w[a] * detJ;
        }
    }

    for (auto i = 0; i < 3 * element_size; ++i)
        for (auto j = i; j < 3 * element_size; ++j)
        {
            if (fabs(K_e_current(i, j)) < epsilon)
                K_e_current(i, j) = 0;
            if (i != j)
                K_e_current(j, i) = K_e_current(i, j);
        }

    for (auto i = 0; i < element_size; ++i)
        for (auto j = i; j < element_size; ++j)
        {
            if (fabs(L_e_current(i, j)) < epsilon)
                L_e_current(i, j) = 0;
            if (i != j)
                L_e_current(j, i) = L_e_current(i, j);
        }

    for (auto i = 0; i < 3 * element_size; ++i)
        for (auto j = 0; j < element_size; ++j)
            if (fabs(S_e_current(i, j)) < epsilon)
                S_e_current(i, j) = 0;

    for (auto i = 0; i < element_size * 3; ++i)
    {
        for (auto j = 0; j < element_size * 3; ++j)
            K_current(i, j) = K_e_current(i, j);
        for (auto j = element_size * 3; j < element_size * 4; ++j)
            K_current(i, j) = 0.;
    }

    for (auto i = 0; i < element_size; ++i)
        for (auto j = 0; j < element_size; ++j)
            L_current(i, j) = L_e_current(i, j);

    for (auto i = 0; i < element_size * 3; ++i)
        for (auto j = 0; j < element_size; ++j)
            S_current(i, j) = S_e_current(i, j);
}
