#include "Graph.h"

#include "wProblem.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"

#include <utility>

#include <Kokkos_UnorderedMap.hpp>
#include <Kokkos_Sort.hpp>

#include "Tpetra_CrsMatrix.hpp"
#include <MatrixMarket_Tpetra.hpp>

#include "LinearSolver.h"

using namespace katropton;
using namespace discrete_problem;

int min(int a, int b)
{
    return (a <= b) ? a : b;
}

int max(int a, int b)
{
    return (a >= b) ? a : b;
}

Graph::Graph()
{
}

Graph::Graph(Teuchos::RCP<Map> map, Teuchos::RCP<Elements_list> elements_list)
{
    using std::pair;
    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;
    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef local_ordinal_type LO;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;
    typedef Tpetra::CrsGraph<local_ordinal_type, global_ordinal_type> crs_graph_type;

    auto numMyElems = elements_list->Get_Element_Number();

    auto numPrimalDPN = map->numPrimalDPN;

    //LinearSolver::getTimers()["graph 1"].start();

    int numLclNodes = map->mapNodesWO->getNodeNumElements();
    int numLclRows = numPrimalDPN * numLclNodes;

    Kokkos::View<size_t *> rowCounts("row counts", numLclRows);

    const int maxConnections = 30;
    Kokkos::UnorderedMap<pair<int, int>, bool> nodenode(numLclNodes * maxConnections);

    size_t numLclEntries = 0;
    Kokkos::parallel_reduce(numMyElems,
                            KOKKOS_LAMBDA(const LO e, size_t &curNumLclEntries) {
                                const int element_size = elements_list->Get_Element_Size(e);
                                for (auto n1 = 0; n1 < element_size; ++n1)
                                    for (auto n2 = n1; n2 < element_size; ++n2)
                                    {
                                        const LO lclNode1 = elements_list->Get_Element_Node(e, n1);
                                        const LO lclNode2 = elements_list->Get_Element_Node(e, n2);

                                        const pair<int, int> key(min(lclNode1, lclNode2), max(lclNode1, lclNode2));

                                        auto result = nodenode.insert(key);

                                        if (result.success())
                                            for (auto i = 0; i < numPrimalDPN; ++i)
                                                for (auto j = 0; j < numPrimalDPN; ++j)
                                                {
                                                    const LO lclRow = numPrimalDPN * lclNode1 + i;
                                                    const LO lclCol = numPrimalDPN * lclNode2 + j;

                                                    if (lclRow < numLclRows)
                                                    {
                                                        if (!(i == 3 && j != 3))
                                                        {
                                                            Kokkos::atomic_fetch_add(&rowCounts(lclRow), 1);
                                                            curNumLclEntries++;
                                                        }
                                                    }
                                                    if (lclNode1 != lclNode2 && lclCol < numLclRows)
                                                    {
                                                        if (!(j == 3 && i != 3))
                                                        {
                                                            Kokkos::atomic_fetch_add(&rowCounts(lclCol), 1);
                                                            curNumLclEntries++;
                                                        }
                                                    }
                                                }
                                    }
                            },
                            numLclEntries);

    Kokkos::View<size_t *> rowOffsets("row offsets", numLclRows + 1);
    Kokkos::parallel_scan(numLclRows,
                          KOKKOS_LAMBDA(int irow, int &update, bool final) {
                              // parallel scan is a multi-pass parallel pattern
                              // In the ‘final’ pass ‘update’ has the prefix value
                              if (final)
                                  rowOffsets(irow) = update;
                              update += rowCounts(irow);
                              if (final && numLclRows == irow + 1)
                                  rowOffsets(irow + 1) = update; // total non-zeros
                          });

    //std::cout<<rowOffsets(numLclRows)<< std::endl;

    //std::cout<<numLclEntries<< std::endl;

    Kokkos::deep_copy(rowCounts, static_cast<size_t>(0));

    Kokkos::View<LO *> colIndices("column indices", rowOffsets(numLclRows));

    Kokkos::parallel_for(nodenode.capacity(),
                         KOKKOS_LAMBDA(int ientry) {
                             if (nodenode.valid_at(ientry))
                             {
                                 const pair<int, int> key = nodenode.key_at(ientry);
                                 const int lclNode1 = key.first;
                                 const int lclNode2 = key.second;

                                 for (auto i = 0; i < numPrimalDPN; ++i)
                                     for (auto j = 0; j < numPrimalDPN; ++j)
                                     {
                                         const LO lclRow = numPrimalDPN * lclNode1 + i;
                                         const LO lclCol = numPrimalDPN * lclNode2 + j;

                                         if (lclRow < numLclRows)
                                         {
                                             if (!(i == 3 && j != 3))
                                             {
                                                 const size_t count = Kokkos::atomic_fetch_add(&rowCounts(lclRow), 1);
                                                 colIndices(rowOffsets(lclRow) + count) = lclCol;
                                             }
                                         }

                                         if (lclNode1 != lclNode2 && lclCol < numLclRows)
                                         {
                                             if (!(j == 3 && i != 3))
                                             {
                                                 const size_t count = Kokkos::atomic_fetch_add(&rowCounts(lclCol), 1);
                                                 colIndices(rowOffsets(lclCol) + count) = lclRow;
                                             }
                                         }
                                     }
                             }
                         });

    //Kokkos::parallel_for( numLclRows ,
    //    KOKKOS_LAMBDA ( int lclRow ) {

    LinearSolver::getTimers()["graph: sort"].start();
    //for (auto lclRow = 0; lclRow < numLclRows; ++lclRow)

    Kokkos::parallel_for(numLclRows,
                         KOKKOS_LAMBDA(int lclRow) {
                             //{
                             auto currentColIndices = subview(colIndices, Kokkos::make_pair(rowOffsets(lclRow), rowOffsets(lclRow + 1)));
                             LO *const lclColIndsRaw = currentColIndices.data();
                             std::sort(lclColIndsRaw, lclColIndsRaw + rowOffsets(lclRow + 1) - rowOffsets(lclRow));
                             //Kokkos::sort(currentColIndices,true);
                             //}
                         });
    LinearSolver::getTimers()["graph: sort"].stop();
    //LinearSolver::getTimers()["graph 1"].stop();
    /*
    {
        std::ofstream outputfile;
        outputfile.open ("colIndices.txt");


        for (auto i=0; i<rowOffsets(numLclRows); ++i)
            outputfile << colIndices(i) << std::endl;

        outputfile.close();
    }
*/

    //    RCP<crs_graph_type> B = rcp(new crs_graph_type (map->mapDofsWO, 0));

    /*
    const int maxConnections = 30;
    Kokkos::View<LO**> nodeConnections ("node connections", numLclNodes,maxConnections+1);
    Kokkos::deep_copy(nodeConnections, 0);

    size_t numLclEntries = 0;
    Kokkos::parallel_reduce (numMyElems,
        KOKKOS_LAMBDA (const LO e, size_t& curNumLclEntries) {

        const int element_size = elements_list->Get_Element_Size(e);
        for ( auto n1 = 0; n1 < element_size; ++n1)
            for ( auto n2 = 0; n2 < element_size; ++n2)
            {
                const LO lclNode1 = elements_list->Get_Element_Node(e,n1);
                const LO lclNode2 = elements_list->Get_Element_Node(e,n2);

                bool newConnection = true;

                for (auto k = 0; k < nodeConnections(lclNode1,0); ++k )
                {
                    if ( nodeConnections(lclNode1,k+1) == lclNode2)
                    {
                        newConnection = false;
                        continue;
                    }
                }

                if (newConnection)
                {
                    if (nodeConnections(lclNode1,0) == maxConnections)
                        std::cout << "error"<<std::endl;

                    Kokkos::atomic_fetch_add (&nodeConnections(lclNode1,0), 1);
                    Kokkos::atomic_fetch_add (&nodeConnections(lclNode1,nodeConnections(lclNode1,0)), lclNode2);

                    for ( auto i = 0; i < numPrimalDPN; ++i)
                        for ( auto j = 0; j < numPrimalDPN; ++j)
                        {
                            if(i==3 && j !=3)
                                continue;

                            const LO lclRow = numPrimalDPN*lclNode1+i;

                            Kokkos::atomic_fetch_add (&rowCounts(lclRow), 1);
                            curNumLclEntries++;
                        }
                }
            }


        }, numLclEntries );

   */

    /*
    rowCounts(numLclRows+1) = numLclEntries;

    std::cout<<rowCounts(numLclRows+1)<< std::endl;

    Kokkos::View<size_t*> colIndices ("column indices", rowCounts(numLclRows+1));

    Kokkos::parallel_for (numMyElems,
        KOKKOS_LAMBDA (const LO e) {

        const int element_size = elements_list->Get_Element_Size(e);
        for ( auto n1 = 0; n1 < element_size; ++n1)
            for ( auto n2 = 0; n2 < element_size; ++n2)
            {
                const LO lclNode1 = elements_list->Get_Element_Node(e,n1);
                const LO lclNode2 = elements_list->Get_Element_Node(e,n2);

                for ( auto i = 0; i < numPrimalDPN; ++i)
                    for ( auto j = 0; j < numPrimalDPN; ++j)
                    {
                        if(i==3 && j !=3)
                            continue;

                        const LO lclRow = numPrimalDPN*lclNode1+i;
                    }
            }
        });

*/
    /*
    RCP<crs_graph_type> B = rcp(new crs_graph_type (map->mapDofsWO, map->mapDofsWO, rowOffsets, colIndices));
    B->fillComplete ();

    AWO = rcp(new crs_graph_type (map->mapDofsWO, 0));
    //LinearSolver::getTimers()["graph 2"].start();
    for(auto e = 0; e < numMyElems; ++e)
    //Kokkos::parallel_for (numMyElems, KOKKOS_LAMBDA (const int e)
    {
        const int element_size = elements_list->Get_Element_Size(e);
        for ( auto n1 = 0; n1 < element_size; ++n1)
            for ( auto n2 = 0; n2 < element_size; ++n2)
                for ( auto i = 0; i < numPrimalDPN; ++i)
                    for ( auto j = 0; j < numPrimalDPN; ++j)
                    {
                        if(i==3 && j !=3)
                            continue;

                        AWO->insertGlobalIndices(numPrimalDPN*(map->mapNodesWO->getGlobalElement(elements_list->Get_Element_Node(e,n1)))+i,
                        tuple<global_ordinal_type> (numPrimalDPN*(map->mapNodesWO->getGlobalElement(elements_list->Get_Element_Node(e,n2)))+j));
                    }
    }
    //);

*/
    //AWO->fillComplete ();

    //AWO = rcp(new crs_graph_type (map->mapDofsWO, B->getColMap(), rowOffsets, colIndices));
    AWO = rcp(new crs_graph_type(map->mapDofsWO, map->mapDofsWO, rowOffsets, colIndices));
    AWO->fillComplete();

    //LinearSolver::getTimers()["graph 2"].stop();

    //    std::cout << "a" << std::endl;
    /*
    typedef Tpetra::CrsMatrix<double,local_ordinal_type,global_ordinal_type> crs_matrix_type;

    Tpetra::MatrixMarket::Writer< crs_matrix_type >::writeSparseGraphFile(std::string("AWO.txt"),AWO);
    Tpetra::MatrixMarket::Writer< crs_matrix_type >::writeSparseGraphFile(std::string("B.txt"),B);

    Tpetra::MatrixMarket::Writer< crs_matrix_type >::writeMapFile (std::string("mapDofsWO.txt"),*(map->mapDofsWO));
    Tpetra::MatrixMarket::Writer< crs_matrix_type >::writeMapFile (std::string("AWO_colmap->txt"),*(AWO->getColMap()));
    Tpetra::MatrixMarket::Writer< crs_matrix_type >::writeMapFile (std::string("B_colmap->txt"),*(B->getColMap()));
*/
    // Create a Tpetra vector for the right hand side (with overlaps).

    const Tpetra::Export<> exportAWO(map->mapDofsWO, map->mapDofs);

    A = rcp(new crs_graph_type(map->mapDofs, 0));
    A->doExport(*AWO, exportAWO, Tpetra::INSERT);
    A->fillComplete();
}
