#include "Map.h"

#include "wProblem.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"
#include "wGroup.h"
#include "wTag.h"

#include <mpi.h>

#include "wContact.h"

#include <Teuchos_Array.hpp>

using namespace katropton;
using namespace discrete_problem;
using namespace tbox;

void std_vector_intersection(std::vector<Tpetra::Vector<>::global_ordinal_type> vec1, std::vector<int> vec2, std::vector<int> &vec3)
{
    std::sort(vec2.begin(), vec2.end());
    for (auto i = 0; i < vec1.size(); ++i)
    {
        if (std::find(vec2.begin(), vec2.end(), vec1[i]) != vec2.end())
            vec3.push_back(vec1[i]);
    }
}

Map::Map()
{
}

Map::Map(Problem &pbl, size_t _numPrimalDPN)
{
    using Teuchos::Array;
    using Teuchos::ArrayRCP;
    using Teuchos::ArrayView;
    using Teuchos::RCP;
    using Teuchos::rcp;

    numPrimalDPN = _numPrimalDPN;

    RCP<const Teuchos::Comm<int>> comm = rcp(new Teuchos::MpiComm<int>(pbl.comm));

    //indexBase = 0;

    int numtasks;
    MPI_Comm_size(MPI_COMM_WORLD, &numtasks);

    auto msh = pbl.msh;

    const Tpetra::global_size_t numNodes = msh->nodes.size();
    const Tpetra::global_size_t numPrimalDOFs = numNodes * numPrimalDPN;

    // Nodes map with overlaps
    // Create a map which associates to each mpi process the nodes included
    // in their elements partition. This map has overlaps associated to the
    // interfaces of the partitions.
    std::vector<global_ordinal_type> myNodesWO = {};
    std::vector<global_ordinal_type> tmpMyNodesWO = {};
    std::vector<global_ordinal_type> myElems = {};

    if (numtasks == 1)
    {
        for (auto e : msh->elems)
            myElems.push_back(e->no - 1);

        for (auto n : msh->nodes)
            myNodesWO.push_back(n->row);
    }
    else
    {
        for (auto e : msh->elems)
        {
            if (e->type() == ELTYPE::HEX8 || e->type() == ELTYPE::TETRA4)
                for (auto n1 : e->nodes)
                    tmpMyNodesWO.push_back(n1->row);
            myElems.push_back(e->no - 1);
        }
        std::sort(tmpMyNodesWO.begin(), tmpMyNodesWO.end());

        myNodesWO.push_back(tmpMyNodesWO[0]);

        for (auto i = 1; i < tmpMyNodesWO.size(); ++i)
            if (tmpMyNodesWO[i] != tmpMyNodesWO[i - 1])
                myNodesWO.push_back(tmpMyNodesWO[i]);
    }

    mapNodesWO = rcp(new tpetra_map_type(numNodes, myNodesWO, indexBase, comm));
    mapElems = Tpetra::createNonContigMap<local_ordinal_type, global_ordinal_type>(myElems, comm);
#if 0
    {
        std::ofstream outputfile;
        std::string name = "mapNodesWO";
        outputfile.open (name + "_" + std::to_string(comm->getRank()) + ".txt");
        //outputfile << mapNodesWO->isOneToOne() << std::endl;  
        for (auto i = 0; i < numNodes; ++i)
        {
            outputfile << mapNodesWO->isNodeGlobalElement(i) << std::endl;
        }
        outputfile.close();
    }
#endif

    // Create a map of the degrees of fredom associated to the nodes of the
    // previous map.
    std::vector<global_ordinal_type> myDofsWO = {};
    for (auto i = 0; i < myNodesWO.size(); ++i)
        for (auto j = 0; j < numPrimalDPN; ++j)
            myDofsWO.push_back(numPrimalDPN * myNodesWO[i] + j);

    mapDofsWO = rcp(new tpetra_map_type(numPrimalDOFs, myDofsWO, indexBase, comm));

    // Loop on the contact and remove from myNodesWO any nodes included in a contact boundary for which the
    // current mpi process has no contact element.
    // This prevent Moertel of not finding the nodes.
    int contacts_number = pbl.Contacts.size();
    if (numtasks != 1)
        for (auto i = 0; i < contacts_number; ++i)
        {
            {
                std::vector<int> v3;
                std_vector_intersection(myNodesWO, pbl.Contacts[i]->slave_boundary_nodes, v3);
                //for (auto j = 0; j < v3.size(); ++j)
                //{
                //    if (comm->getRank()==2)
                //    std::cout << "Proc " <<comm->getRank() << " interface " << i << " "<< j << "/" << v3.size() << " " << v3[j] << std::endl;
                //}
                //if (comm->getRank()==2)
                //std::cout << "Proc " <<comm->getRank() << " interface " << i << " AT MOST slave nodes to remove " << v3.size() << std::endl;
                for (auto j = 0; j < pbl.Contacts[i]->slave_elems.size(); ++j)
                    for (auto jj = 0; jj < pbl.Contacts[i]->slave_elems[j]->nodes.size(); ++jj)
                    {
                        auto no = pbl.Contacts[i]->slave_elems[j]->nodes[jj]->row;
                        //if (comm->getRank()==2)
                        //    std::cout << "Proc " <<comm->getRank() << " interface " << i << " elem " << pbl.Contacts[i]->slave_elems[j]->no << " node " <<  no << std::endl;
                        auto p = std::find(v3.begin(), v3.end(), no);
                        if (p != v3.end())
                        {
                            v3.erase(p);
                        }
                    }
                //if (comm->getRank()==2)
                //std::cout << "Proc " <<comm->getRank() << " interface " << i << " slave nodes to remove " << v3.size() << std::endl;
                for (auto j = 0; j < v3.size(); ++j)
                {
                    //if (comm->getRank()==2)
                    //std::cout << "Proc " <<comm->getRank() << " interface " << i << " "<< j << "/" << v3.size() << " " << v3[j] << std::endl;
                    auto p = std::find(myNodesWO.begin(), myNodesWO.end(), v3[j]);
                    if (p != myNodesWO.end())
                    {
                        //std::cout<< comm->getRank() <<" erase slave: " << *p << std::endl;
                        myNodesWO.erase(p);
                    }
                }
            }
            {
                std::vector<int> v3;
                std_vector_intersection(myNodesWO, pbl.Contacts[i]->master_boundary_nodes, v3);
                //if (comm->getRank()==2)
                //std::cout << "Proc " <<comm->getRank() << " interface " << i << " master " << v3.size() << std::endl;
                for (auto j = 0; j < pbl.Contacts[i]->master_elems.size(); ++j)
                    for (auto jj = 0; jj < pbl.Contacts[i]->master_elems[j]->nodes.size(); ++jj)
                    {
                        auto no = pbl.Contacts[i]->master_elems[j]->nodes[jj]->row;
                        auto p = std::find(v3.begin(), v3.end(), no);
                        if (p != v3.end())
                        {
                            v3.erase(p);
                        }
                    }
                for (auto j = 0; j < v3.size(); ++j)
                {
                    auto p = std::find(myNodesWO.begin(), myNodesWO.end(), v3[j]);
                    if (p != myNodesWO.end())
                    {
                        //std::cout<< comm->getRank() <<" erase master: " << *p << std::endl;
                        myNodesWO.erase(p);
                    }
                }
            }
        }
    RCP<const Tpetra::Map<>> mapNodesWO_contact = rcp(new tpetra_map_type(numNodes, myNodesWO, indexBase, comm));

    // Create a map of nodes without overlaps. Each node is owned by one and
    // only mpi process.
    mapNodes = Tpetra::createOneToOne(mapNodesWO_contact);

#if 0
    {
        std::ofstream outputfile;
        std::string name = "mapNodes";
        outputfile.open (name + "_" + std::to_string(comm->getRank()) + ".txt");
        //outputfile << mapNodes->isOneToOne() << std::endl;  
        for (auto i = 0; i < numNodes; ++i)
        {
            outputfile << mapNodesWO->isNodeGlobalElement(i) << " " << mapNodesWO_contact->isNodeGlobalElement(i) << " " << mapNodes->isNodeGlobalElement(i) << std::endl;
        }
        outputfile.close();
    }
#endif

    //std::cout << "Proc " << comm->getRank() << " 342 " << mapNodes->isNodeGlobalElement (342) << mapNodesWO_contact->isNodeGlobalElement (342) <<std::endl;

    // Create a map of the degrees of fredom associated to the nodes of the
    // previous map.
    const size_t numMyNodes = mapNodes->getNodeNumElements();
    Teuchos::ArrayView<const global_ordinal_type> myNodes = mapNodes->getNodeElementList();

    Array<global_ordinal_type> myDofs(numMyNodes * numPrimalDPN);
    for (auto i = 0; i < numMyNodes; ++i)
        for (auto j = 0; j < numPrimalDPN; ++j)
            myDofs[numPrimalDPN * i + j] = numPrimalDPN * myNodes[i] + j;

    mapDofs = rcp(new tpetra_map_type(numPrimalDOFs, myDofs, indexBase, comm));

    // Construct a Map that puts all the equations on processor 0 (for outputs).
    const size_t myRank = comm->getRank();
    const size_t numLclIndices = (myRank == 0) ? numPrimalDOFs : 0;

    mapOutput = rcp(new tpetra_map_type(numPrimalDOFs, numLclIndices, indexBase, comm));
}
Map &Map::operator=(const Map &rhs)
{
    mapNodesWO = rhs.mapNodesWO;
    mapNodes = rhs.mapNodes;
    mapElems = rhs.mapElems;
    mapDofsWO = rhs.mapDofsWO;
    mapDofs = rhs.mapDofs;
    mapOutput = rhs.mapOutput;
    numPrimalDPN = rhs.numPrimalDPN;
    return *this;
};
