#ifndef WRANDOMFIELD_H
#define WRANDOMFIELD_H

#define RandomField_DEBUG

#include "Stokhos_Sacado_Kokkos_MP_Vector.hpp"
#include "Stokhos_KL_ExponentialRandomField.hpp"
#include "EnsembleTraits.h"

template <typename Scalar, typename Device>
class RandomField
{
public:
  typedef Kokkos::View<Scalar *, Kokkos::LayoutLeft, Device> RandomVariableView;
  typedef typename RandomVariableView::size_type size_type;

  typedef typename Teuchos::ScalarTraits<Scalar>::coordinateType MeshScalar;
  typedef Stokhos::KL::ExponentialRandomField<MeshScalar, Device> rf_type;
  typedef Teuchos::ScalarTraits<MeshScalar> MST;
  typedef typename EnsembleTraits<Scalar>::value_type mean_type;

  bool isRandom = false;
  rf_type m_rf;            // Exponential random field
  RandomVariableView m_rv; // KL random variables
  mean_type g;             // Mean
  mean_type delta;         // Dispersion level
  mean_type a, b;
  int num_rv;
  int ndim = 3; // Dimension of the physical domain

  RandomField(Teuchos::RCP<Teuchos::ParameterList> randomParams, Scalar *m_rv_values)
  {
    isRandom = randomParams->get("Is random", false);
    g = randomParams->get("Mean", MST::one());
    delta = randomParams->get("Dispersion level", MST::zero());
    const MeshScalar BeginX = randomParams->get("Begin X", MST::zero());
    const MeshScalar BeginY = randomParams->get("Begin Y", MST::zero());
    const MeshScalar BeginZ = randomParams->get("Begin Z", MST::zero());
    const MeshScalar EndX = randomParams->get("End X", MST::one());
    const MeshScalar EndY = randomParams->get("End Y", MST::one());
    const MeshScalar EndZ = randomParams->get("End Z", MST::one());
    num_rv = randomParams->get("Number random variables", 0);
    const MeshScalar correlation_lengthX = randomParams->get("Correlation length X", MST::one());
    const MeshScalar correlation_lengthY = randomParams->get("Correlation length Y", MST::one());
    const MeshScalar correlation_lengthZ = randomParams->get("Correlation length Z", MST::one());

    if (correlation_lengthZ == 0)
      ndim = 2;

    const int num_KL_X = randomParams->get("Number of KL Terms X", 0);
    const int num_KL_Y = randomParams->get("Number of KL Terms Y", 0);
    const int num_KL_Z = randomParams->get("Number of KL Terms Z", 0);

    mean_type eps = randomParams->get("Bound Perturbation Size", 1e-6);
    mean_type tol = randomParams->get("Nonlinear Solver Tolerance", 1e-10);
    int max_it = randomParams->get("Maximum Nonlinear Solver Iterations", 100);

    if (isRandom)
    {
      Teuchos::ParameterList solverParams;
      solverParams.set("Number of KL Terms", num_rv);
      solverParams.set("Mean", MST::zero());
      solverParams.set("Standard Deviation", MST::one());
      solverParams.set("Bound Perturbation Size", eps);
      solverParams.set("Nonlinear Solver Tolerance", tol);
      solverParams.set("Maximum Nonlinear Solver Iterations", max_it);
      Teuchos::Array<MeshScalar> domain_upper(ndim, 1.0),
          domain_lower(ndim, 0.0),
          correlation_lengths(ndim, correlation_lengthX);
      Teuchos::Array<int> num_KL_per_dim(ndim, 1.0);

      domain_upper[0] = EndX;
      domain_upper[1] = EndY;
      if (ndim == 3)
        domain_upper[2] = EndZ;

      domain_lower[0] = BeginX;
      domain_lower[1] = BeginY;
      if (ndim == 3)
        domain_lower[2] = BeginZ;

      correlation_lengths[0] = correlation_lengthX;
      correlation_lengths[1] = correlation_lengthY;
      if (ndim == 3)
        correlation_lengths[2] = correlation_lengthZ;

      if (num_KL_X != 0 && num_KL_Y != 0 && num_KL_Z != 0)
      {
        num_KL_per_dim[0] = num_KL_X;
        num_KL_per_dim[1] = num_KL_Y;
        if (ndim == 3)
          num_KL_per_dim[2] = num_KL_Z;
        solverParams.set("Number of KL Terms per dimension", num_KL_per_dim);
      }

      solverParams.set("Domain Upper Bounds", domain_upper);
      solverParams.set("Domain Lower Bounds", domain_lower);
      solverParams.set("Correlation Lengths", correlation_lengths);

      m_rf = rf_type(solverParams);
      m_rv = RandomVariableView("KL Random Variables", num_rv);
      for (int i = 0; i < num_rv; ++i)
      {
        m_rv(i) = m_rv_values[i];
        for (int ell = 0; ell < EnsembleTraits<Scalar>::size; ++ell)
        {
          if (!std::isfinite(EnsembleTraits<Scalar>::coeff(m_rv(i), ell)))
          {
#ifdef RandomField_DEBUG
            printf("Warning 1: random variable of index %d for sample %d is not finite: %f;\n", i, ell, EnsembleTraits<Scalar>::coeff(m_rv(i), ell));
#endif
            EnsembleTraits<Scalar>::coeff(m_rv(i), ell) = 0;
          }
        }
      }
      a = log(g / (sqrt(1 + delta * delta)));
      b = sqrt(log(1 + delta * delta));
    }
  }

  KOKKOS_INLINE_FUNCTION
  Scalar operator()(const MeshScalar x, const MeshScalar y, const MeshScalar z) const
  {
    if (isRandom)
    {
      Scalar Xi;
      if (ndim == 3)
      {
        const MeshScalar point[3] = {x, y, z};
        Xi = m_rf.evaluate(point, m_rv);
      }
      else
      {
        const MeshScalar point[2] = {x, y};
        Xi = m_rf.evaluate(point, m_rv);
      }
      for (int ell = 0; ell < EnsembleTraits<Scalar>::size; ++ell)
      {
        if (!std::isfinite(EnsembleTraits<Scalar>::coeff(Xi, ell)))
        {
#ifdef RandomField_DEBUG
          const MeshScalar point2[3] = {0, 0, 0};
          Scalar Xi2 = m_rf.evaluate(point2, m_rv);

          std::string m_rv_string;
          for (int i = 0; i < num_rv; ++i)
            m_rv_string += std::to_string(EnsembleTraits<Scalar>::coeff(m_rv(i), ell)) + ' ';
          printf("Warning 2: Xi (= %f) (Xi(0,0,0) = %f) is not finite for sample %d, point position [%f,%f,%f] random variables [%s];\n", EnsembleTraits<Scalar>::coeff(Xi, ell), EnsembleTraits<Scalar>::coeff(Xi2, ell), ell, x, y, z, m_rv_string.c_str());
#endif
          EnsembleTraits<Scalar>::coeff(Xi, ell) = 0;
        }
      }

      Scalar val = exp(a + b * Xi);

      for (int ell = 0; ell < EnsembleTraits<Scalar>::size; ++ell)
      {
        if (!std::isfinite(EnsembleTraits<Scalar>::coeff(val, ell)) || EnsembleTraits<Scalar>::coeff(val, ell) <= 0.)
        {
#ifdef RandomField_DEBUG
          std::string m_rv_string;
          for (int i = 0; i < num_rv; ++i)
            m_rv_string += std::to_string(EnsembleTraits<Scalar>::coeff(m_rv(i), ell)) + ' ';
          printf("Warning 3: val (= %f) is not finite or zero for sample %d with Xi = %f, point position [%f,%f,%f] random variables [%s];\n", EnsembleTraits<Scalar>::coeff(val, ell), ell, EnsembleTraits<Scalar>::coeff(Xi, ell), x, y, z, m_rv_string.c_str());
#endif
          EnsembleTraits<Scalar>::coeff(val, ell) = exp(a);
        }
      }

      return val;
    }
    else
      return Teuchos::ScalarTraits<Scalar>::one();
  }
};

#endif //WRANDOMFIELD_H
