#ifndef WSOURCES_LIST_H
#define WSOURCES_LIST_H

#include "katropton.h"

#include "Map.h"
#include "wSource.h"
#include "wTag.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Kokkos_ViewFactory.hpp>

namespace katropton
{
namespace discrete_problem
{
template <typename scalar>
class Sources_list
{
private:
    Kokkos::View<int **, Kokkos::LayoutRight> sources_elements_list;
    Kokkos::View<scalar ***, Kokkos::LayoutRight> sources_values_list;

    Tpetra::Vector<>::local_ordinal_type sources_number;
    Tpetra::Vector<>::local_ordinal_type sources_size;
    Tpetra::Vector<>::local_ordinal_type sources_elements;

    Tpetra::Vector<>::local_ordinal_type sources_time;
    Tpetra::Vector<>::local_ordinal_type sources_values;

public:
    Sources_list();
    Sources_list(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<Elements_list> elements_list);
    inline local_ordinal_type getSourceNumber() { return sources_number; };
    inline local_ordinal_type getSourceSize(local_ordinal_type i)
    {
        return sources_elements_list(i, sources_size);
    };
    inline local_ordinal_type getSourceElement(local_ordinal_type i, local_ordinal_type n1)
    {
        return sources_elements_list(i, sources_elements + n1);
    };
    inline scalar getSourceValue(local_ordinal_type i)
    {
        return sources_values_list(i, 0, sources_values);
    };
};

template <typename scalar>
Sources_list<scalar>::Sources_list()
{
    sources_number = 0;
    sources_size = 0;
    sources_elements = 0;

    sources_time = 0;
    sources_values = 0;

    sources_elements_list = Kokkos::View<int **, Kokkos::LayoutRight>("R", sources_number, 0);
    sources_values_list = Kokkos::View<scalar ***, Kokkos::LayoutRight>("R", sources_number, 0, 0);
}

template <typename scalar>
Sources_list<scalar>::Sources_list(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<Elements_list> elements_list)
{
    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    sources_number = pbl.Sources.size();
    local_ordinal_type maxNumElemsPerSource = 0;

    std::vector<std::vector<global_ordinal_type>> SourceElems = {};

    //const size_t myRank = map->mapElems->getComm()->getRank();

    for (auto i = 0; i < sources_number; ++i)
    {
        std::vector<global_ordinal_type> mySourceElems = {};

        //std::cout << "myRank " << myRank << " source number " << i << " maximum source size " << pbl.Sources[i]->tag->elems.size() << std::endl;

        for (auto j = 0; j < pbl.Sources[i]->tag->elems.size(); ++j)
        {
            local_ordinal_type e = map->mapElems->getLocalElement(pbl.Sources[i]->tag->elems[j]->no - 1);

            //std::cout << "myRank " << myRank << " j " << j << " element no -1 " << pbl.Sources[i]->tag->elems[j]->no-1 << " e " << e << std::endl;

            if (elements_list->Get_Element_Type(e) == static_cast<int>(tbox::ELTYPE::TETRA4))
                ;
            else if (elements_list->Get_Element_Type(e) == static_cast<int>(tbox::ELTYPE::HEX8))
                ;
            else
                continue;

            //std::cout << "myRank " << myRank << " j " << j << " e " << e << " OK" << std::endl;

            mySourceElems.push_back(e);
        }
        //std::cout << "myRank " << myRank << " source number " << i << " actual source size " << mySourceElems.size() << std::endl;
        SourceElems.push_back(mySourceElems);
        if (maxNumElemsPerSource < mySourceElems.size())
            maxNumElemsPerSource = mySourceElems.size();
    }

    sources_size = 0;
    sources_elements = 1;
    sources_elements_list = Kokkos::View<int **, Kokkos::LayoutRight>("R", sources_number, sources_elements + maxNumElemsPerSource);

    for (auto i = 0; i < sources_number; ++i)
    {
        sources_elements_list(i, sources_size) = SourceElems[i].size();

        for (auto j = 0; j < SourceElems[i].size(); ++j)
            sources_elements_list(i, sources_elements + j) = SourceElems[i][j];
    }

    sources_time = 0;
    sources_values = 1;
    sources_values_list = Kokkos::View<scalar ***, Kokkos::LayoutRight>("R", sources_number, 1, 2);
    for (auto i = 0; i < sources_number; ++i)
    {
        sources_values_list(i, 0, sources_time) = 0.; //to change latter

        sources_values_list(i, 0, sources_values) = pbl.Sources[i]->value;
    }
}
}; // namespace discrete_problem
}; // namespace katropton

#endif //WSOURCES_LIST_H
