#ifndef mtMatrix_H
#define mtMatrix_H

#include "Stokhos_Tpetra_MP_Vector.hpp"

template <class T, int rows>
class tVector;

template <class T, int rows, int cols>
class tMatrix
{
public:
    tMatrix()
    {
        for (auto i = 0; i < rows; ++i)
            for (auto j = 0; j < cols; ++j)
                data[i * cols + j] = (T)0;
    }
    T &operator()(int row, int col);
    tMatrix operator+(tMatrix rhs);
    void operator+=(tMatrix rhs);
    T det();       //only for 3x3
    tMatrix inv(); //only for 3x3
    tMatrix operator*(tMatrix rhs);
    tVector<T, rows> operator*(tVector<T, cols> rhs);
    template <class U>
    tVector<typename Sacado::Promote<T, U>::type, rows> operator*(tVector<U, cols> rhs);
    void print();
    void clean()
    {
        for (auto i = 0; i < rows; ++i)
            for (auto j = 0; j < cols; ++j)
                data[i * cols + j] = (T)0;
    }

private:
    T data[rows * cols];
    int numRows = rows;
    int numCols = cols;
    int getRows() { return numRows; }
    int getCols() { return numCols; }
};

template <class T, int rows, int cols>
template <class U>
tVector<typename Sacado::Promote<T, U>::type, rows> tMatrix<T, rows, cols>::operator*(tVector<U, cols> rhs)
{
    tVector<typename Sacado::Promote<T, U>::type, rows> results;
    for (auto i = 0; i < rows; ++i)
        for (auto j = 0; j < cols; ++j)
            results(i) += data[i * cols + j] * rhs(j);
    return results;
}

template <class T, int rows, int cols>
tMatrix<T, rows, cols> tMatrix<T, rows, cols>::operator+(tMatrix<T, rows, cols> rhs)
{
    tMatrix<T, rows, cols> results;
    for (auto i = 0; i < rows; ++i)
        for (auto j = 0; j < cols; ++j)
            results(i, j) = data[i * cols + j] + rhs(i, j);
    return results;
}

template <class T, int rows, int cols>
void tMatrix<T, rows, cols>::operator+=(tMatrix<T, rows, cols> rhs)
{
    for (auto i = 0; i < rows; ++i)
        for (auto j = 0; j < cols; ++j)
            data[i * cols + j] += rhs(i, j);
}

template <class T, int rows, int cols>
tMatrix<T, rows, cols> tMatrix<T, rows, cols>::operator*(tMatrix<T, rows, cols> rhs)
{
    tMatrix<T, rows, cols> results;
    for (auto i = 0; i < rows; ++i)
        for (auto j = 0; j < rows; ++j)
            for (auto k = 0; k < rows; ++k)
                results(i, j) += data[i * cols + k] * rhs(k, j);
    return results;
}

template <class T, int rows, int cols>
T &tMatrix<T, rows, cols>::operator()(int row, int col)
{
    return data[row * cols + col];
}

template <class T, int rows, int cols>
T tMatrix<T, rows, cols>::det()
{
    return data[0 * cols + 0] * (data[1 * cols + 1] * data[2 * cols + 2] - data[1 * cols + 2] * data[2 * cols + 1]) - data[0 * cols + 1] * (data[1 * cols + 0] * data[2 * cols + 2] - data[1 * cols + 2] * data[2 * cols + 0]) + data[0 * cols + 2] * (data[1 * cols + 0] * data[2 * cols + 1] - data[1 * cols + 1] * data[2 * cols + 0]);
}

template <class T, int rows, int cols>
tMatrix<T, rows, cols> tMatrix<T, rows, cols>::inv()
{
    tMatrix<T, rows, cols> results;
    T det = this->det();
    results(0, 0) = (1. / det) * (data[1 * cols + 1] * data[2 * cols + 2] - data[1 * cols + 2] * data[2 * cols + 1]);
    results(0, 1) = (1. / det) * (data[0 * cols + 2] * data[2 * cols + 1] - data[2 * cols + 2] * data[0 * cols + 1]);
    results(0, 2) = (1. / det) * (data[0 * cols + 1] * data[1 * cols + 2] - data[1 * cols + 1] * data[0 * cols + 2]);

    results(1, 0) = (1. / det) * (data[1 * cols + 2] * data[2 * cols + 0] - data[2 * cols + 2] * data[1 * cols + 0]);
    results(1, 1) = (1. / det) * (data[0 * cols + 0] * data[2 * cols + 2] - data[2 * cols + 0] * data[0 * cols + 2]);
    results(1, 2) = (1. / det) * (data[0 * cols + 2] * data[1 * cols + 0] - data[1 * cols + 2] * data[0 * cols + 0]);

    results(2, 0) = (1. / det) * (data[1 * cols + 0] * data[2 * cols + 1] - data[2 * cols + 0] * data[1 * cols + 1]);
    results(2, 1) = (1. / det) * (data[0 * cols + 1] * data[2 * cols + 0] - data[2 * cols + 1] * data[0 * cols + 0]);
    results(2, 2) = (1. / det) * (data[0 * cols + 0] * data[1 * cols + 1] - data[1 * cols + 0] * data[0 * cols + 1]);
    return results;
}

template <class T, int rows, int cols>
tVector<T, rows> tMatrix<T, rows, cols>::operator*(tVector<T, cols> rhs)
{
    tVector<T, rows> results;
    for (auto i = 0; i < rows; ++i)
        for (auto j = 0; j < cols; ++j)
            results(i) += data[i * cols + j] * rhs(j);
    return results;
}

template <class T, int rows, int cols>
void tMatrix<T, rows, cols>::print()
{
    for (auto i = 0; i < rows; ++i)
    {
        for (auto j = 0; j < cols; ++j)
            std::cout << data[i * cols + j] << " ";
        std::cout << std::endl;
    }
}

template <class T, int rows>
class tVector
{
public:
    tVector()
    {
        for (auto i = 0; i < rows; ++i)
            data[i] = (T)0;
    }
    T &operator()(int row);
    tVector operator+(tVector rhs);
    void operator+=(tVector rhs);
    T dotproduct(tVector rhs);
    template <class U>
    typename Sacado::Promote<T, U>::type dotproduct(tVector<U, rows> rhs);
    tMatrix<T, rows, rows> ddotproduct(tVector rhs);
    void print();
    void clean()
    {
        for (auto i = 0; i < rows; ++i)
            data[i] = (T)0;
    }

private:
    T data[rows];
    int numRows = rows;
    int getRows() { return numRows; }
};

template <class T, int rows>
T &tVector<T, rows>::operator()(int row)
{
    return data[row];
}

template <class T, int rows>
tVector<T, rows> tVector<T, rows>::operator+(tVector<T, rows> rhs)
{
    tVector<T, rows> results;
    for (auto i = 0; i < rows; ++i)
        results(i) = data[i] + rhs(i);
    return results;
}

template <class T, int rows>
void tVector<T, rows>::operator+=(tVector<T, rows> rhs)
{
    for (auto i = 0; i < rows; ++i)
        data[i] += rhs(i);
}

template <class T, int rows>
T tVector<T, rows>::dotproduct(tVector<T, rows> rhs)
{
    T result = (T)0;
    for (auto i = 0; i < rows; ++i)
        result += data[i] * rhs(i);
    return result;
}

template <class T, int rows>
template <class U>
typename Sacado::Promote<T, U>::type tVector<T, rows>::dotproduct(tVector<U, rows> rhs)
{
    typename Sacado::Promote<T, U>::type result = 0;
    for (auto i = 0; i < rows; ++i)
        result += data[i] * rhs(i);
    return result;
}

template <class T, int rows>
tMatrix<T, rows, rows> tVector<T, rows>::ddotproduct(tVector<T, rows> rhs)
{
    tMatrix<T, rows, rows> results;
    for (auto i = 0; i < rows; ++i)
        for (auto j = 0; j < rows; ++j)
            results(i, j) = data[i] * rhs(j);
    return results;
}

template <class T, int rows>
void tVector<T, rows>::print()
{
    for (auto i = 0; i < rows; ++i)
        std::cout << data[i] << std::endl;
}

#endif
