#include <Kokkos_Core.hpp>
#include "wCacheHex8.h"
#include "wCacheQuad4.h"
#include "wCacheTetra4.h"
#include "wCacheTri3.h"
#include "tMatrix.h"
#include "Sacado.hpp"

#include "EnsembleTraits.h"

tbox::CacheTri3 &trilinosTri3GetCache();
double trilinosTri3BuildJ(int k, Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes);

tbox::CacheQuad4 &trilinosQuad4GetCache();
double trilinosQuad4BuildJ(int k, Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes);

tbox::CacheHex8 &trilinosHex8GetCache();
double trilinosHex8BuildJ(int k, tMatrix<double, 3, 3> &J, Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes);

tbox::CacheTetra4 &trilinosTetra4GetCache();
double trilinosTetra4BuildJ(int k, tMatrix<double, 3, 3> &J, Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes);

template <typename scalar>
Kokkos::View<scalar *, Kokkos::LayoutRight> get_material_properties(Kokkos::View<scalar ***, Kokkos::LayoutRight> mesh_Materials, scalar T, int curent_material)
{
    size_t materials_T = 0;

    typedef EnsembleTraits<scalar> ET;
    const int ensemble_size = ET::size;

    Kokkos::View<scalar *, Kokkos::LayoutRight> material_properties("A", mesh_Materials.extent(2));
    auto material_properties_size = material_properties.extent(0);

    for (int j = 0; j < ensemble_size; ++j)
    {
        typename ET::value_type Tj = ET::coeff(T, j);

        size_t index_0 = 0;
        size_t index_1 = mesh_Materials.extent(1) - 1;
        if (Tj <= mesh_Materials(curent_material, index_0, materials_T))
            for (auto i = 0; i < material_properties_size; ++i)
                material_properties(i) = mesh_Materials(curent_material, index_0, i);
        else if (Tj > mesh_Materials(curent_material, index_1, materials_T))
            for (auto i = 0; i < material_properties_size; ++i)
                material_properties(i) = mesh_Materials(curent_material, index_1, i);
        else
        {
            while (index_1 != index_0 + 1)
            {
                size_t index_c = index_0 + (index_1 - index_0) / 2;

                if (Tj == mesh_Materials(curent_material, index_c, materials_T))
                {
                    index_1 = index_c;
                    index_0 = index_1 - 1;
                }
                else if (Tj < mesh_Materials(curent_material, index_c, materials_T))
                    index_1 = index_c;
                else
                    index_0 = index_c;
            }

            typename ET::value_type coeff = (Tj - mesh_Materials(curent_material, index_0, materials_T)) / (mesh_Materials(curent_material, index_1, materials_T) - mesh_Materials(curent_material, index_0, materials_T));
            for (auto i = 0; i < material_properties_size; ++i)
                ET::coeff(material_properties(i), j) = mesh_Materials(curent_material, index_0, i) + coeff * (mesh_Materials(curent_material, index_1, i) - mesh_Materials(curent_material, index_0, i));
        }
    }

    return material_properties;
}

template <typename scalar>
Kokkos::View<scalar *, Kokkos::LayoutRight> get_material_properties(Kokkos::View<scalar ***, Kokkos::LayoutRight> mesh_Materials, Sacado::Fad::SFad<scalar, 32> T, int curent_material)
{
    size_t materials_T = 0;

    typedef EnsembleTraits<scalar> ET;
    const int ensemble_size = ET::size;

    Kokkos::View<scalar *, Kokkos::LayoutRight> material_properties("A", mesh_Materials.extent(2));
    auto material_properties_size = material_properties.extent(0);

    for (int j = 0; j < ensemble_size; ++j)
    {
        typename ET::value_type Tj = ET::coeff(T.val(), j);

        size_t index_0 = 0;
        size_t index_1 = mesh_Materials.extent(1) - 1;

        if (Tj <= mesh_Materials(curent_material, index_0, materials_T))
            for (auto i = 0; i < material_properties_size; ++i)
                material_properties(i) = mesh_Materials(curent_material, index_0, i);
        else if (Tj > mesh_Materials(curent_material, index_1, materials_T))
            for (auto i = 0; i < material_properties_size; ++i)
                material_properties(i) = mesh_Materials(curent_material, index_1, i);
        else
        {
            while (index_1 != index_0 + 1)
            {
                size_t index_c = index_0 + (index_1 - index_0) / 2;

                if (Tj == mesh_Materials(curent_material, index_c, materials_T))
                {
                    index_1 = index_c;
                    index_0 = index_1 - 1;
                }
                else if (Tj < mesh_Materials(curent_material, index_c, materials_T))
                    index_1 = index_c;
                else
                    index_0 = index_c;
            }

            scalar coeff = (Tj - mesh_Materials(curent_material, index_0, materials_T)) / (mesh_Materials(curent_material, index_1, materials_T) - mesh_Materials(curent_material, index_0, materials_T));
            for (auto i = 0; i < material_properties_size; ++i)
                ET::coeff(material_properties(i), j) = mesh_Materials(curent_material, index_0, i) + coeff * (mesh_Materials(curent_material, index_1, i) - mesh_Materials(curent_material, index_0, i));
        }
    }

    return material_properties;
}

// K matrix
template <typename scalar>
Kokkos::View<scalar **, Kokkos::LayoutRight> trilinosTetra4BuildK_mechanic(Kokkos::View<scalar ***, Kokkos::LayoutRight> mesh_Materials, Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes)
{
    typedef Kokkos::View<scalar **, Kokkos::LayoutRight> matrix_view_type;
    typedef Kokkos::View<scalar *, Kokkos::LayoutRight> material_properties_type;

    size_t elems_type = 0;
    size_t elems_size = 1;
    size_t elems_media = 2;
    size_t elems_nodes = 3;

    size_t materials_T = 0;
    size_t materials_E = 1;
    size_t materials_nu = 2;
    size_t materials_k = 3;
    size_t materials_beta = 4;

    tbox::CacheTetra4 &cache = trilinosTetra4GetCache();
    tbox::GaussTetra4 &gauss = cache.gauss;

    matrix_view_type K_e("A", 3 * currentElem(elems_size), 3 * currentElem(elems_size));
    tVector<double, 3> factor1, factor2, dffi, dffj;
    tVector<scalar, 3> factor3;
    tMatrix<double, 3, 3> J, Jinv;
    tMatrix<scalar, 3, 3> TMP1;
    tMatrix<scalar, 9, 9> H;
    scalar E, nu, lambda, G;
    // Machine epsilon

    double epsilon = 2e-16;

    for (auto a = 0; a < gauss.p.size(); ++a)
    {
        double detJ = trilinosTetra4BuildJ(a, J, currentElem, mesh_Nodes);
        Jinv = J.inv();

        for (auto i = 0; i < currentElem(elems_size); ++i)
        {
            for (auto k = 0; k < 3; ++k)
                dffi(k) = cache.dff[a][i][k];

            factor1 = Jinv * dffi;

            for (auto j = i; j < currentElem(elems_size); ++j)
            {
                material_properties_type material_properties = get_material_properties<scalar>(mesh_Materials, (scalar)0, currentElem(elems_media));
                E = material_properties(materials_E);   //mesh_Materials(currentElem(elems_media),0,materials_E);
                nu = material_properties(materials_nu); //mesh_Materials(currentElem(elems_media),0,materials_nu);
                lambda = (nu * E) / ((1 + nu) * (1 - 2 * nu));
                G = E / (2 * (1 + nu));

                // H matrix
                H.clean();

                for (auto i = 0; i < 3; ++i)
                    for (auto j = 0; j < 3; ++j)
                        for (auto k = 0; k < 3; ++k)
                            for (auto l = 0; l < 3; ++l)
                            {
                                if (i == j && k == l)
                                    H(i * 3 + j, k * 3 + l) = lambda;
                                if (i == k && j == l)
                                    H(i * 3 + j, k * 3 + l) = H(i * 3 + j, k * 3 + l) + G;
                                if (i == l && j == k)
                                    H(i * 3 + j, k * 3 + l) = H(i * 3 + j, k * 3 + l) + G;
                            }

                for (auto k = 0; k < 3; ++k)
                    dffj(k) = cache.dff[a][j][k];

                factor2 = Jinv * dffj;

                TMP1.clean();
                for (auto k = 0; k < 3; ++k)
                    for (auto m = 0; m < 3; ++m)
                    {
                        for (auto l = 0; l < 3; ++l)
                            for (auto n = 0; n < 3; ++n)
                                TMP1(l, n) = H(k * 3 + l, m * 3 + n);

                        factor3 = TMP1 * factor2;

                        scalar dotProduct = factor1.dotproduct(factor3);

                        K_e(i * 3 + k, j * 3 + m) += dotProduct * gauss.w[a] * detJ;
                    }
            }
        }
    }

    for (auto i = 0; i < 3 * currentElem(elems_size); ++i)
        for (auto j = i; j < 3 * currentElem(elems_size); ++j)
        {
            if (fabs(K_e(i, j)) < epsilon)
                K_e(i, j) = 0;
            if (i != j)
                K_e(j, i) = K_e(i, j);
        }
    return K_e;
}

template <typename scalar>
Kokkos::View<scalar **, Kokkos::LayoutRight> trilinosHex8BuildK_mechanic(Kokkos::View<scalar ***, Kokkos::LayoutRight> mesh_Materials, Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes)
{
    typedef Kokkos::View<scalar **, Kokkos::LayoutRight> matrix_view_type;
    typedef Kokkos::View<scalar *, Kokkos::LayoutRight> material_properties_type;
    typedef Sacado::Fad::SFad<scalar, 32> fad_scalar_type;

    size_t elems_type = 0;
    size_t elems_size = 1;
    size_t elems_media = 2;
    size_t elems_nodes = 3;

    size_t materials_T = 0;
    size_t materials_E = 1;
    size_t materials_nu = 2;
    size_t materials_k = 3;
    size_t materials_beta = 4;

    tbox::CacheHex8 &cache = trilinosHex8GetCache();
    tbox::GaussHex8 &gauss = cache.gauss;

    tVector<scalar, 32> tValues;

    fad_scalar_type val[32];

    for (auto i = 0; i < 32; ++i)
    {
        val[i].val() = tValues(i);
        val[i].diff(i, 32);
    }
    fad_scalar_type val_at_gauss[8];

    matrix_view_type K_e("A", 3 * currentElem(elems_size), 3 * currentElem(elems_size));
    matrix_view_type out_K_e("A", 3 * currentElem(elems_size), 4 * currentElem(elems_size));
    tVector<double, 3> factor1, factor2, dffi, dffj;
    tVector<scalar, 3> factor3;
    tMatrix<double, 3, 3> J, Jinv;
    tMatrix<scalar, 3, 3> TMP1;
    tMatrix<scalar, 9, 9> H;
    scalar E, nu, lambda, G;
    // Machine epsilon

    double epsilon = 2e-16;

    // Looping on gauss points
    for (auto a = 0; a < gauss.p.size(); ++a)
    {
        val_at_gauss[a] = val[24] * cache.ff[a][0];
        for (auto j = 1; j < currentElem(elems_size); ++j)
            val_at_gauss[a] += val[24 + j] * cache.ff[a][j];

        material_properties_type material_properties = get_material_properties<scalar>(mesh_Materials, val_at_gauss[a], currentElem(elems_media));
        E = material_properties(materials_E);   //mesh_Materials(currentElem(elems_media),0,materials_E);
        nu = material_properties(materials_nu); //mesh_Materials(currentElem(elems_media),0,materials_nu);
        lambda = (nu * E) / ((1 + nu) * (1 - 2 * nu));
        G = E / (2 * (1 + nu));

        // H matrix
        H.clean();

        for (auto i = 0; i < 3; ++i)
            for (auto j = 0; j < 3; ++j)
                for (auto k = 0; k < 3; ++k)
                    for (auto l = 0; l < 3; ++l)
                    {
                        if (i == j && k == l)
                            H(i * 3 + j, k * 3 + l) = lambda;
                        if (i == k && j == l)
                            H(i * 3 + j, k * 3 + l) = H(i * 3 + j, k * 3 + l) + G;
                        if (i == l && j == k)
                            H(i * 3 + j, k * 3 + l) = H(i * 3 + j, k * 3 + l) + G;
                    }

        double detJ = trilinosHex8BuildJ(a, J, currentElem, mesh_Nodes);
        Jinv = J.inv();

        for (auto i = 0; i < currentElem(elems_size); ++i)
        {
            for (auto k = 0; k < 3; ++k)
                dffi(k) = cache.dff[a][i][k];

            factor1 = Jinv * dffi;

            for (auto j = i; j < currentElem(elems_size); ++j)
            {
                for (auto k = 0; k < 3; ++k)
                    dffj(k) = cache.dff[a][j][k];

                factor2 = Jinv * dffj;

                TMP1.clean();
                for (auto k = 0; k < 3; ++k)
                    for (auto m = 0; m < 3; ++m)
                    {
                        for (auto l = 0; l < 3; ++l)
                            for (auto n = 0; n < 3; ++n)
                                TMP1(l, n) = H(k * 3 + l, m * 3 + n);

                        factor3 = TMP1 * factor2;

                        scalar dotProduct = factor1.dotproduct(factor3);

                        K_e(i * 3 + k, j * 3 + m) += dotProduct * gauss.w[a] * detJ;
                    }
            }
        }
    }

    for (auto i = 0; i < 3 * currentElem(elems_size); ++i)
        for (auto j = i; j < 3 * currentElem(elems_size); ++j)
        {
            if (fabs(K_e(i, j)) < epsilon)
                K_e(i, j) = 0;
            if (i != j)
                K_e(j, i) = K_e(i, j);
        }

    fad_scalar_type elem_res[24];

    for (auto i = 0; i < 24; ++i)
        for (auto j = 0; j < 24; ++j)
            elem_res[i] += K_e(i, j) * val[j];

    for (auto i = 0; i < 24; ++i)
        for (auto j = 0; j < 32; ++j)
            out_K_e(i, j) = elem_res[i].fastAccessDx(j);

    return out_K_e;
}

// L matrix
template <typename scalar>
Kokkos::View<scalar **, Kokkos::LayoutRight> trilinosTetra4BuildL_thermic(Kokkos::View<scalar ***, Kokkos::LayoutRight> mesh_Materials, Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes)
{
    typedef Kokkos::View<scalar **, Kokkos::LayoutRight> matrix_view_type;
    typedef Kokkos::View<scalar *, Kokkos::LayoutRight> material_properties_type;

    size_t elems_type = 0;
    size_t elems_size = 1;
    size_t elems_media = 2;
    size_t elems_nodes = 3;

    size_t materials_T = 0;
    size_t materials_E = 1;
    size_t materials_nu = 2;
    size_t materials_k = 3;
    size_t materials_beta = 4;

    tbox::CacheTetra4 &cache = trilinosTetra4GetCache();
    tbox::GaussTetra4 &gauss = cache.gauss;

    matrix_view_type L_e("A", currentElem(elems_size), currentElem(elems_size));
    tVector<double, 3> factor1, factor3, dffi, dffj;
    tVector<scalar, 3> factor2;
    tMatrix<double, 3, 3> J, Jinv;
    scalar conductivity;
    tMatrix<scalar, 3, 3> K_conductivity;
    // Machine epsilon

    double epsilon = 2e-16;

    for (auto a = 0; a < gauss.p.size(); ++a)
    {
        double detJ = trilinosTetra4BuildJ(a, J, currentElem, mesh_Nodes);
        Jinv = J.inv();

        for (auto i = 0; i < currentElem(elems_size); ++i)
        {
            material_properties_type material_properties = get_material_properties<scalar>(mesh_Materials, (scalar)0, currentElem(elems_media));
            conductivity = material_properties(materials_k); //mesh_Materials(currentElem(elems_media),0,materials_k);
            K_conductivity.clean();
            for (auto i = 0; i < 3; ++i)
                K_conductivity(i, i) = conductivity;

            for (auto k = 0; k < 3; ++k)
                dffi(k) = cache.dff[a][i][k];
            factor1 = Jinv * dffi;
            factor2 = K_conductivity * factor1;

            for (auto j = i; j < currentElem(elems_size); ++j)
            {
                for (auto k = 0; k < 3; ++k)
                    dffj(k) = cache.dff[a][j][k];
                factor3 = Jinv * dffj;
                scalar dotProduct = factor3.dotproduct(factor2);
                L_e(i, j) += dotProduct * gauss.w[a] * detJ;
            }
        }
    }

    for (auto i = 0; i < currentElem(elems_size); ++i)
        for (auto j = i; j < currentElem(elems_size); ++j)
        {
            if (fabs(L_e(i, j)) < epsilon)
                L_e(i, j) = 0;
            if (i != j)
                L_e(j, i) = L_e(i, j);
        }
    return L_e;
}

template <typename scalar>
Kokkos::View<scalar **, Kokkos::LayoutRight> trilinosHex8BuildL_thermic(Kokkos::View<scalar ***, Kokkos::LayoutRight> mesh_Materials, Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes)
{
    typedef Kokkos::View<scalar **, Kokkos::LayoutRight> matrix_view_type;
    typedef Kokkos::View<scalar *, Kokkos::LayoutRight> material_properties_type;

    size_t elems_type = 0;
    size_t elems_size = 1;
    size_t elems_media = 2;
    size_t elems_nodes = 3;

    size_t materials_T = 0;
    size_t materials_E = 1;
    size_t materials_nu = 2;
    size_t materials_k = 3;
    size_t materials_beta = 4;

    tbox::CacheHex8 &cache = trilinosHex8GetCache();
    tbox::GaussHex8 &gauss = cache.gauss;

    matrix_view_type L_e("A", currentElem(elems_size), currentElem(elems_size));
    tVector<double, 3> factor1, factor3, dffi, dffj;
    tVector<scalar, 3> factor2;
    tMatrix<double, 3, 3> J, Jinv;
    scalar conductivity;
    tMatrix<scalar, 3, 3> K_conductivity;
    // Machine epsilon

    double epsilon = 2e-16;

    for (auto a = 0; a < gauss.p.size(); ++a)
    {
        double detJ = trilinosHex8BuildJ(a, J, currentElem, mesh_Nodes);
        Jinv = J.inv();

        for (auto i = 0; i < currentElem(elems_size); ++i)
        {
            material_properties_type material_properties = get_material_properties<scalar>(mesh_Materials, (scalar)0, currentElem(elems_media));
            conductivity = material_properties(materials_k); //mesh_Materials(currentElem(elems_media),0,materials_k);
            K_conductivity.clean();
            for (auto i = 0; i < 3; ++i)
                K_conductivity(i, i) = conductivity;

            for (auto k = 0; k < 3; ++k)
                dffi(k) = cache.dff[a][i][k];
            factor1 = Jinv * dffi;
            factor2 = K_conductivity * factor1;

            for (auto j = i; j < currentElem(elems_size); ++j)
            {
                for (auto k = 0; k < 3; ++k)
                    dffj(k) = cache.dff[a][j][k];
                factor3 = Jinv * dffj;
                scalar dotProduct = factor3.dotproduct(factor2);
                L_e(i, j) += dotProduct * gauss.w[a] * detJ;
            }
        }
    }

    for (auto i = 0; i < currentElem(elems_size); ++i)
        for (auto j = i; j < currentElem(elems_size); ++j)
        {
            if (fabs(L_e(i, j)) < epsilon)
                L_e(i, j) = 0;
            if (i != j)
                L_e(j, i) = L_e(i, j);
        }
    return L_e;
}

// S matrix
template <typename scalar>
Kokkos::View<scalar **, Kokkos::LayoutRight> trilinosTetra4BuildS_thermomechanic(Kokkos::View<scalar ***, Kokkos::LayoutRight> mesh_Materials, Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes)
{
    typedef Kokkos::View<scalar **, Kokkos::LayoutRight> matrix_view_type;
    typedef Kokkos::View<scalar *, Kokkos::LayoutRight> material_properties_type;

    size_t elems_type = 0;
    size_t elems_size = 1;
    size_t elems_media = 2;
    size_t elems_nodes = 3;

    size_t materials_T = 0;
    size_t materials_E = 1;
    size_t materials_nu = 2;
    size_t materials_k = 3;
    size_t materials_beta = 4;

    tbox::CacheTetra4 &cache = trilinosTetra4GetCache();
    tbox::GaussTetra4 &gauss = cache.gauss;

    matrix_view_type S_e("A", 3 * currentElem(elems_size), currentElem(elems_size)); // dimension: 24*8
    tVector<double, 3> factor1, factor3, dffi;
    tVector<double, 4> ff;
    tVector<scalar, 3> factor2;
    tMatrix<double, 3, 3> J, Jinv;
    scalar E, nu, lambda, G, beta;
    tMatrix<scalar, 3, 3> D;
    // Machine epsilon

    double epsilon = 2e-16;

    for (auto a = 0; a < gauss.p.size(); ++a)
    {
        double detJ = trilinosTetra4BuildJ(a, J, currentElem, mesh_Nodes);
        Jinv = J.inv();

        for (auto j = 0; j < currentElem(elems_size); ++j)
            ff(j) = cache.ff[a][j];

        for (auto i = 0; i < currentElem(elems_size); ++i)
        {
            material_properties_type material_properties = get_material_properties<scalar>(mesh_Materials, (scalar)0, currentElem(elems_media));
            E = material_properties(materials_E);       //mesh_Materials(currentElem(elems_media),0,materials_E);
            nu = material_properties(materials_nu);     //mesh_Materials(currentElem(elems_media),0,materials_nu);
            beta = material_properties(materials_beta); //mesh_Materials(currentElem(elems_media),0,materials_beta);
            lambda = (nu * E) / ((1 + nu) * (1 - 2 * nu));
            G = E / (2 * (1 + nu));

            // D matrix
            D.clean();
            for (auto i = 0; i < 3; ++i)
                D(i, i) = -3 * (lambda + (2. / 3.) * G) * beta;

            for (auto k = 0; k < 3; ++k)
                dffi(k) = cache.dff[a][i][k];
            factor1 = Jinv * dffi;
            factor2 = D * factor1;

            for (auto j = 0; j < currentElem(elems_size); ++j)
                for (auto k = 0; k < 3; ++k)
                    S_e(i * 3 + k, j) += factor2(k) * ff(j) * gauss.w[a] * detJ;
        }
    }

    for (auto i = 0; i < 3 * currentElem(elems_size); ++i)
        for (auto j = 0; j < currentElem(elems_size); ++j)
            if (fabs(S_e(i, j)) < epsilon)
                S_e(i, j) = 0;

    return S_e;
}

template <typename scalar>
Kokkos::View<scalar **, Kokkos::LayoutRight> trilinosHex8BuildS_thermomechanic(Kokkos::View<scalar ***, Kokkos::LayoutRight> mesh_Materials, Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes)
{
    typedef Kokkos::View<scalar **, Kokkos::LayoutRight> matrix_view_type;
    typedef Kokkos::View<scalar *, Kokkos::LayoutRight> material_properties_type;

    size_t elems_type = 0;
    size_t elems_size = 1;
    size_t elems_media = 2;
    size_t elems_nodes = 3;

    size_t materials_T = 0;
    size_t materials_E = 1;
    size_t materials_nu = 2;
    size_t materials_k = 3;
    size_t materials_beta = 4;

    tbox::CacheHex8 &cache = trilinosHex8GetCache();
    tbox::GaussHex8 &gauss = cache.gauss;

    matrix_view_type S_e("A", 3 * currentElem(elems_size), currentElem(elems_size)); // dimension: 24*8
    tVector<double, 3> factor1, factor3, dffi;
    tVector<double, 8> ff;
    tVector<scalar, 3> factor2;
    tMatrix<double, 3, 3> J, Jinv;
    scalar E, nu, lambda, G, beta;
    tMatrix<scalar, 3, 3> D;
    // Machine epsilon

    double epsilon = 2e-16;

    for (auto a = 0; a < gauss.p.size(); ++a)
    {
        double detJ = trilinosHex8BuildJ(a, J, currentElem, mesh_Nodes);
        Jinv = J.inv();

        for (auto j = 0; j < currentElem(elems_size); ++j)
            ff(j) = cache.ff[a][j];

        for (auto i = 0; i < currentElem(elems_size); ++i)
        {
            material_properties_type material_properties = get_material_properties<scalar>(mesh_Materials, (scalar)0, currentElem(elems_media));
            E = material_properties(materials_E);
            nu = material_properties(materials_nu);
            beta = material_properties(materials_beta);
            lambda = (nu * E) / ((1 + nu) * (1 - 2 * nu));
            G = E / (2 * (1 + nu));

            // D matrix
            D.clean();
            for (auto i = 0; i < 3; ++i)
                D(i, i) = -3 * (lambda + (2. / 3.) * G) * beta;

            for (auto k = 0; k < 3; ++k)
                dffi(k) = cache.dff[a][i][k];
            factor1 = Jinv * dffi;
            factor2 = D * factor1;

            for (auto j = 0; j < currentElem(elems_size); ++j)
                for (auto k = 0; k < 3; ++k)
                    S_e(i * 3 + k, j) += factor2(k) * ff(j) * gauss.w[a] * detJ;
        }
    }

    for (auto i = 0; i < 3 * currentElem(elems_size); ++i)
        for (auto j = 0; j < currentElem(elems_size); ++j)
            if (fabs(S_e(i, j)) < epsilon)
                S_e(i, j) = 0;

    return S_e;
}

template <typename scalar>
Kokkos::View<scalar **, Kokkos::LayoutRight> trilinosQuad4BuildN(Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes)
{
    typedef Kokkos::View<scalar **, Kokkos::LayoutRight> matrix_view_type;

    size_t elems_type = 0;
    size_t elems_size = 1;
    size_t elems_media = 2;
    size_t elems_nodes = 3;

    tbox::CacheQuad4 &cache = trilinosQuad4GetCache();
    tbox::GaussQuad4 &gauss = cache.gauss;

    matrix_view_type N_e("A", currentElem(elems_size), currentElem(elems_size)); // dimension: 4*4
    tVector<double, 4> ff;
    // Machine epsilon

    double epsilon = 2e-16;

    for (auto a = 0; a < gauss.p.size(); ++a)
    {
        double detJ = trilinosQuad4BuildJ(a, currentElem, mesh_Nodes);
        //std::cout  << detJ << std::endl;
        //Jinv = J.inv();

        for (auto j = 0; j < currentElem(elems_size); ++j)
            ff(j) = cache.ff[a][j];

        for (auto i = 0; i < currentElem(elems_size); ++i)
            for (auto j = 0; j < currentElem(elems_size); ++j)
                N_e(i, j) += ff(i) * ff(j) * gauss.w[a] * detJ;
    }

    //    for  ( auto i = 0; i < currentElem(elems_size); ++i )
    //        for ( auto j = 0; j < currentElem(elems_size); ++j )
    //            if ( fabs(N_e(i,j)) < epsilon )
    //                N_e(i,j) = 0;

    return N_e;
}

template <typename scalar>
Kokkos::View<scalar **, Kokkos::LayoutRight> trilinosTri3BuildN(Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes)
{
    typedef Kokkos::View<scalar **, Kokkos::LayoutRight> matrix_view_type;

    size_t elems_type = 0;
    size_t elems_size = 1;
    size_t elems_media = 2;
    size_t elems_nodes = 3;

    tbox::CacheTri3 &cache = trilinosTri3GetCache();
    tbox::GaussTri3 &gauss = cache.gauss;

    matrix_view_type N_e("A", currentElem(elems_size), currentElem(elems_size)); // dimension: 4*4
    tVector<double, 4> ff;
    // Machine epsilon

    double epsilon = 2e-16;

    for (auto a = 0; a < gauss.p.size(); ++a)
    {
        double detJ = trilinosTri3BuildJ(a, currentElem, mesh_Nodes);
        //std::cout  << detJ << std::endl;
        //Jinv = J.inv();

        for (auto j = 0; j < currentElem(elems_size); ++j)
            ff(j) = cache.ff[a][j];

        for (auto i = 0; i < currentElem(elems_size); ++i)
            for (auto j = 0; j < currentElem(elems_size); ++j)
                N_e(i, j) += ff(i) * ff(j) * gauss.w[a] * detJ;
    }

    //    for  ( auto i = 0; i < currentElem(elems_size); ++i )
    //        for ( auto j = 0; j < currentElem(elems_size); ++j )
    //            if ( fabs(N_e(i,j)) < epsilon )
    //                N_e(i,j) = 0;

    return N_e;
}

template <typename scalar>
Kokkos::View<scalar **, Kokkos::LayoutRight> trilinosHex8BuildN(Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes)
{
    typedef Kokkos::View<scalar **, Kokkos::LayoutRight> matrix_view_type;

    size_t elems_type = 0;
    size_t elems_size = 1;
    size_t elems_media = 2;
    size_t elems_nodes = 3;

    tMatrix<double, 3, 3> J;

    tbox::CacheHex8 &cache = trilinosHex8GetCache();
    tbox::GaussHex8 &gauss = cache.gauss;

    matrix_view_type N_e("A", currentElem(elems_size), currentElem(elems_size)); // dimension: 4*4
    tVector<double, 4> ff;
    // Machine epsilon

    double epsilon = 2e-16;

    for (auto a = 0; a < gauss.p.size(); ++a)
    {
        double detJ = trilinosHex8BuildJ(a, J, currentElem, mesh_Nodes);
        //std::cout  << detJ << std::endl;
        //Jinv = J.inv();

        for (auto j = 0; j < currentElem(elems_size); ++j)
            ff(j) = cache.ff[a][j];

        for (auto i = 0; i < currentElem(elems_size); ++i)
            for (auto j = 0; j < currentElem(elems_size); ++j)
                N_e(i, j) += ff(i) * ff(j) * gauss.w[a] * detJ;
    }

    return N_e;
}

template <typename scalar>
Kokkos::View<scalar **, Kokkos::LayoutRight> trilinosTetra4BuildN(Kokkos::View<int *, Kokkos::LayoutLeft> currentElem, Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes)
{
    typedef Kokkos::View<scalar **, Kokkos::LayoutRight> matrix_view_type;

    size_t elems_type = 0;
    size_t elems_size = 1;
    size_t elems_media = 2;
    size_t elems_nodes = 3;

    tMatrix<double, 3, 3> J;

    tbox::CacheTetra4 &cache = trilinosTetra4GetCache();
    tbox::GaussTetra4 &gauss = cache.gauss;

    matrix_view_type N_e("A", currentElem(elems_size), currentElem(elems_size)); // dimension: 4*4
    tVector<double, 4> ff;
    // Machine epsilon

    double epsilon = 2e-16;

    for (auto a = 0; a < gauss.p.size(); ++a)
    {
        double detJ = trilinosTetra4BuildJ(a, J, currentElem, mesh_Nodes);
        //std::cout  << detJ << std::endl;
        //Jinv = J.inv();

        for (auto j = 0; j < currentElem(elems_size); ++j)
            ff(j) = cache.ff[a][j];

        for (auto i = 0; i < currentElem(elems_size); ++i)
            for (auto j = 0; j < currentElem(elems_size); ++j)
                N_e(i, j) += ff(i) * ff(j) * gauss.w[a] * detJ;
    }

    return N_e;
}
